#define EmmaCtrl(arg) db 0x82,arg

// BIOS enter point
// 0xf000:0x0000
ORG 0x0

start:
mov ax,0xf000
mov ds,ax
mov eax,0
cpuid  // emma's cpu will return "Emma" in ebx
cmp [emma_magic],ebx // is CPU Emma VM?
jz cpu_ok
dead_loop: jmp dead_loop
cpu_ok:
// write interrupt table with default handler
mov eax,0xf000
mov ss,ax
mov sp,0xfff0
shl eax,16
mov ds,ax
mov ax,default_int
mov ecx,0x100

set_handlers:
     mov [ecx*4-4],eax
     loop set_handlers
// now write BIOS service interrupt handlers

//
//
// BIOS service
mov word [0x10*4],int_10
mov word [0x11*4],int_11
mov word [0x12*4],int_12
mov word [0x13*4],int_13
mov word [0x14*4],int_14
mov word [0x15*4],int_15
mov word [0x16*4],int_16
mov word [0x17*4],int_17
mov word [0x18*4],int_18
mov word [0x19*4],int_19
mov word [0x1a*4],int_1a

mov word [0x1b*4],key_brk
mov word [0x1d*4],video_params
mov word [0x1e*4],fd_params
mov word [0x1f*4],gr_chars
//Init bios data area
data_area:  mov ax,0x40
	    mov ds,ax
	    mov word [0x13],0x280 //640k


mov ax,0xf000
mov ds,ax
mov si,FIRST_MESSAGE
call say
jmp int_19
mov si,DEAD_MSG
call say
jmp dead_loop
// boot_beg
// Read first sector in 0000:0x7c00
// and next give control to it
boot_beg:   mov ax,0
            mov es,ax
            mov bx,0x7c00
            mov ax,0x0201 // al-count ah-02h(read sector)
            mov ch,0 // track
	    mov cl,1 // sector
    	    mov dl,0 // driver 0(0-3 fd 0x80-0x81 hd)
            mov dh,0 // head 0
	    clc
            int 0x13 // read sector
	    jnc disk_ok
	    call disk_error
	    jmp boot_beg
// disk Ok, GO
disk_ok:    mov si,BOOTOK_MSG
	    call say
//	    mov eax,0x0
//	    mov si,END_MSG
//	    EmmaLog
	    mov es,ax
	    mov ds,ax
	    mov ecx,eax
	    mov ebx,eax
	    mov esi,eax
	    jmp 0x0000:0x7c00 // go!!
// disk error, reset and come back	    
disk_error: mov si,DISK_ERROR
	    call say
	    mov ah,0
	    int 0x13
	    ret
// say asciiz string DS:SI
say:	    lodsb
	    mov ah,0xe
	    cmp al,0x0
	    jz say_end
	    int 0x10
	    jmp say
say_end:    ret
//
int_10:
	EmmaCtrl(0x10)
	jmp default_int
int_11:
	EmmaCtrl(0x11)	
	jmp default_int
int_12:
	EmmaCtrl(0x12)	
	jmp default_int
int_13:
	EmmaCtrl(0x13)	
	jmp keep_carry_int
int_14:
	EmmaCtrl(0x14)	
	jmp default_int
int_15:
	EmmaCtrl(0x15)	
	jmp keep_carry_int
int_16:
	EmmaCtrl(0x16)	
	jmp default_int
int_17:
	EmmaCtrl(0x17)	
	jmp default_int
int_18:
	EmmaCtrl(0x18)	
	jmp default_int
int_19:
	jmp boot_beg
//	jmp default_int
int_1a:
	EmmaCtrl(0x1a)
	jmp keep_carry_int
key_brk:
	jmp default_int	
// for handlers that uses carry flag as return value
keep_carry_int:
	      push ax 
	      push bp
	      mov ax,0xfffe      // carry flag clear mask
	      mov bp,sp
	      jnc no_carry
set_carry:    or word [ss:bp+0x8],0x1
	      jmp carry_end
no_carry:     and [ss:bp+0x8],ax // clear carry
carry_end:    pop bp
	      pop ax
// default handler for unhandled interrupts
default_int:
	iret

//fd params table
fd_params:
	db 0,0,0,
	db 2,		//sector size 512 bytes
	db 18,		//18 sectors
	db 0,0,0,0xf6,0,0
//
	db 0,0,0,
	db 2,		//sector size 512 bytes
	db 18,		//18 sectors
	db 0,0,0,0xf6,0,0
	
video_params:
gr_chars:
    	
// data area
FIRST_MESSAGE: db "(Emma Virtual Machine)'es semi-external BIOS version 0.0.1",0xa
COPYRIGHT:     db "Copyright (C) 2003, Vitja Makarov",0xa,0x0
BOOTOK_MSG:    db "Booting from fd0",0xa,0xa,0xa,0x0
DISK_ERROR:    db "Couldn't boot: disk error, trying again...",0xa,0x0
END_MSG:       db "Exiting bios area...",0xa,0x0
DEAD_MSG:      db "Entring dead state...",0xa,0x0
emma_magic:    db "ammE" // is Emma spelled back
