/*
 *  This file is part of Emma Virtual Machine
 *  Copyright (C) 2003 Vitja Makarov <no1sm@narod.ru>
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include "cpu.h"
#include "dis386.h"
#include "newinls.h"

# if defined (MIPS_COUNTER)
#  include <sys/timeb.h>
# endif


CPU::CPU()
{
    /* nullify registers */
    memset((void *) &Regs, 0, sizeof(Regs));
    memset((void *) &SRegs, 0, sizeof(Regs));
    eflags = 0;
    mem = NULL;
    
    /* clear some flags */
    amode = osize = 
    pref_lock = pref_rep = 0;
    
    segment = _zero_;
};


void CPU::set_memory(void *ptr)
{
    mem = (char *) ptr;
}

CPU::~CPU()
{
    if (mem)
	free(mem);
}

#ifdef MIPS_COUNTER

unsigned long 	cmds_num = 0;
static double 	cmds_time = 0;
static int 	can_say_mips = 1;
static double 	tmp_time = 0;
double 		mips = 0;

// returns value:
// time in seconds + milliseconds/1000
static inline double gettime()
{
    struct timeb tp;
    ftime(&tp);
    return (double) tp.time + (double) (tp.millitm / 1000.);
}
#endif

extern void debug_go(CPU *,int);

int CPU::run()
{
    int ret = 0, rrr = 0;

    // if ( we are in protected mode)
    //     goto begin_pm;
    // else
    // goto begin_rm;

 begin_rm:

    sync_code();

    if ( (ret=setjmp(STH_THIS exc_buf)) )
    {	/* exception handling not done yet */
#if defined (DEBUGGER)
      debug_go(STH_TH, ret);
#endif
    }

 next_ins:

#ifdef MIPS_COUNTER
    if (can_say_mips) {
      tmp_time = gettime();
      can_say_mips = 0;
      cmds_num = 0;
      cmds_time = 0;
    }
#endif

#if defined( DEBUGGER )
	debug_go(STH_TH,0);
#endif

	i_go();

#ifdef MIPS_COUNTER
	if ( (can_say_mips = !(++cmds_num % 500000)) ) {
	    tmp_time = gettime() - tmp_time;
	    cmds_time += tmp_time;
	    mips = cmds_num / (cmds_time * 1000000.);
	    printf("MIPS=%.03f\n",mips);
	}
#endif
	rrr = STH_THIS external_check();
	
	if ( rrr >= 0 )
	    return rrr;

    goto next_ins;
}


int CPU::set_handler(NewHandler o16,NewHandler o32, Bit32u opcode)
{
    switch (opcode & 0xff00)
    {
    case OPCODE_ONEBYTE:
	 one_byte_table [opcode & 0xff] [0] = o16;
	 one_byte_table [opcode & 0xff] [1] = o32;
	 return 0;
	 
    case OPCODE_TWOBYTE:
	 two_byte_table [opcode & 0xff] [0] = o16;
	 two_byte_table [opcode & 0xff] [1] = o32;
	 return 0;
	 
    case OPCODE_REGGATE:
	 grp_table [ (opcode >> 3) & 0xff ] [opcode & 0x7 ] [0] = o16;
	 grp_table [ (opcode >> 3) & 0xff ] [opcode & 0x7 ] [1] = o32;
	 return 0;

    /* default: couldn't set */
    }
  return 1;
}
