/*
 *  This file is part of Emma Virtual Machine
 *  Copyright (C) 2003 Vitja Makarov <no1sm@narod.ru>
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef _CPU_H_
#define _CPU_H_

#include <stdio.h>
#include <setjmp.h>

#include "config.h"
#include "disasm.h"

// some fitches and macroses was tooken from DosEmu
#define SEG_ADR(t,seg,off) t(memptr(lineaddr(seg,off)))

/* Some exception numbers + 1 */
#define EXC_DE   1
#define EXC_DB   2
#define EXC_NMI  3
#define EXC_BP   4
#define EXC_OF   5
#define EXC_BR   6
#define EXC_UD   7
#define EXC_NM   8
#define EXC_DF   9
//#define EXC_DF   9
#define EXC_TS   11
#define EXC_NP   12
#define EXC_SS   13
#define EXC_GP   14
#define EXC_PF   15
   
#include "regs.h"

#if 0
inline void set_bit(Bit32u &x,Bit32u mask){
  x|=mask;
}

inline void clear_bit(Bit32u &x,Bit32u mask){
  x&=~mask;
}

inline int get_bit(Bit32u x,Bit32u mask){
  return (x&mask) && 0x1;
}

inline void sc_bit(Bit32u &x,Bit32u mask,int set){// set/clear mask 
  set?set_bit(x,mask):clear_bit(x,mask);
}
#else
# define set_bit(x,mask) ((x)|=(mask))
# define clear_bit(x,mask) ((x)&=~(mask))
# define sc_bit(x,mask,s) ((s)?set_bit(x,mask):clear_bit(x,mask))
# define get_bit(x,mask)  ((x)&(mask) && 0x1)
#endif

class CPU;

    typedef void (STH_WHAT *NewHandler)();

    extern /* const */ NewHandler one_byte_table[0x100][4];
    extern /* const */ NewHandler two_byte_table[0x100][4];
    extern /* const */ NewHandler grp_table[20][8][4];

#if defined (_STH_)
extern CPU __cpu__;
#endif

class CPU{
 public:
  CPU();
  int initMemory(int sz);
  void set_memory(void *);
  virtual ~CPU();

  char *mem;
  char *textMem;

  STH_PREF int run();
  STH_PREF Bit32u 
    lineaddr(Bit16u segment,Bit32u offset)
    {
      return (Bit32u)((segment<<4)+(offset&0xffff))&0xfffff;
    }

STH_PREF  void *memptr(Bit32u addr)
	    {
	        return STH_THIS mem+(addr&0xfffff);
	    }
STH_PREF  void *memptr(Bit16u a,Bit32u b)
	    {
		return STH_THIS mem+(lineaddr(a,b)&0xfffff);
	    }

STH_PREF  inline int stack_push32(Bit32u data);
STH_PREF  inline int stack_push16(Bit16u data);
STH_PREF  inline Bit32u stack_pop32();
STH_PREF  inline Bit16u stack_pop16();

  // flags that can be set automaticly
#if 0  
  inline void set_sf(Bit32u res,int s){
    sc_bit(eflags,SF,(res&sign_bit(s) ) && 0x1);
  }
  inline void set_pf(Bit32u res,int=0){
    sc_bit(eflags,PF,parity_matrix[res&0xff]);
  }
  inline void set_zf(Bit32u res,int s){
    sc_bit(eflags,ZF,!(res&val_mask(s)));
  }
  inline void set_af(Bit32u res,int s){
    sc_bit(eflags,ZF,!(res&val_mask(s)));
  }
  inline void set_szp(Bit32u res,int s){
    set_sf(res,s),set_zf(res,s),set_pf(res,s);
  };
#else

# define set_sf(res,s) sc_bit(__eflags, SF,( (res) &sign_bit(s) ))
# define set_pf(res,s) sc_bit(__eflags, PF, parity_matrix[(res)&0xff])
# define set_zf(res,s) sc_bit(__eflags, ZF, (res)==0 )
# define set_af(res,s) sc_bit(__eflags, AF,!((res)&val_mask(s)) )

#define set_szp(res,s) (set_sf(res,s),set_zf(res,s),set_pf(res,s))

#endif  
  //set/clear
#if 0
  inline void cls(){ clear_bit(eflags,SF);}
  inline void clz(){ clear_bit(eflags,ZF);}
  inline void clo(){ clear_bit(eflags,OF);}
  inline void clc(){ clear_bit(eflags,CF);}
  inline void cla(){ clear_bit(eflags,AF);}
  inline void clp(){ clear_bit(eflags,PF);}
  inline void cld(){ clear_bit(eflags,DF);}
  inline void cli(){ clear_bit(eflags,IF);}

  inline void sts(){ set_bit(eflags,SF);}
  inline void stz(){ set_bit(eflags,ZF);}
  inline void sto(){ set_bit(eflags,OF);}
  inline void stc(){ set_bit(eflags,CF);}
  inline void sta(){ set_bit(eflags,AF);}
  inline void stp(){ set_bit(eflags,PF);}
  inline void std(){ set_bit(eflags,DF);}
  inline void sti(){ set_bit(eflags,IF);}
#endif
  // commands that set flags in dependence of command
  // all these commands have next arguments:
  // o1  -- first operand
  // o2  -- second operand
  // res -- command's result
  // s   -- operandes's size
# define DEF_FLAGS
  
# if !defined (DEF_FLAGS)
STH_PREF  inline void set_flags_add(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_adc(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_sub(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_sbb(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_dec(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_inc(Bit32u o1,Bit32u o2,Bit32u res,int s);

STH_PREF  inline void set_flags_shl(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_shr(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_sal(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_sar(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_rol(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_ror(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_rcl(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_rcr(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_flags_logical(Bit32u o1,Bit32u o2,Bit32u res,int s);

STH_PREF  inline void set_cf_like_add(Bit32u o1,Bit32u o2,Bit32u res,int=0);
STH_PREF  inline void set_of_like_add(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_af_like_add(Bit32u o1,Bit32u o2,Bit32u res,int=0);
  //
STH_PREF  inline void set_cf_like_sub(Bit32u o1,Bit32u o2,Bit32u res,int=0);
STH_PREF  inline void set_of_like_sub(Bit32u o1,Bit32u o2,Bit32u res,int s);
STH_PREF  inline void set_af_like_sub(Bit32u o1,Bit32u o2,Bit32u res,int=0);
# endif  

  Bit32u Regs[9];
  Bit16u SRegs[9];
  Bit32u eflags;

  Bit32u EIP;

#define eip STH_THIS EIP  


// 
// some external stuff...
// these all should be released out of libcpu.a
// as soon as emma isn't virtual machine with it's own interface
// it more seems to be just a library that emulates cpu
//
// external_check - runs after each instruction, to check for 
//			diffrent things that we can call external
//			for CPU that is little box inside big
//			 (like PIC,#NMI,#ERROR)
//			if return value >=0 CPU::run() returns it
//
// e_(in|out)(8|16|32) - port access
//
//
// sync_regs_(down|up) - may be used to release register/(other data) 
//			 exchange between diffrent emulators
//			  	for example dosemu/emma
//
// emmaCtrl	       - could be used to call some internal commands
//				(not specific for native CPU)
//
// set_handler         - set handler function for specified opcode
//			   coded according with that what we want:
//
//			one byte command: just it number (0x0-0xff)
//			    two byte    : OPCODE_TWOBYTE | (0x0-0xff)
//	        modrm gate (opcode extension by modrm): 
//				OPCODE_REGGATE | 
// 					(grp_number | (cmd_number << 3) )
//
//			 this handler should decode his arguments by itself
//				using commands called decode stuff
//				or it's own
//
//		NOTE: you should pass two handlers as arguments
//			for 16-bit operand size and 32-bit,
//			address mode could be taken from CPU::amode
//				segment prefix from      CPU::segment
//				and so go on
//

#define OPCODE_ONEBYTE  0x0000
#define OPCODE_TWOBYTE  0x0100
#define OPCODE_REGGATE  0x0200

STH_PREF inline void sync_regs_up();
STH_PREF inline void sync_regs_down();
STH_PREF Bit8u  e_in8  (Bit32u port);
STH_PREF Bit16u e_in16 (Bit32u port);
STH_PREF Bit32u e_in32 (Bit32u port);
STH_PREF void e_out8   (Bit32u port,Bit8u  src);
STH_PREF void e_out16  (Bit32u port,Bit16u src);
STH_PREF void e_out32  (Bit32u port,Bit32u src);
STH_PREF int external_check();	
STH_PREF void emmaCtrl();
STH_PREF int  set_handler(NewHandler o16,
			        NewHandler o32, Bit32u opcode);

//
// decode stuff
//

STH_PREF inline void loadModRM   ();

STH_PREF inline Bit8u  *countRM	();
STH_PREF inline Bit8u  *loadRM8  ();
STH_PREF inline Bit16u *loadRM16 ();
STH_PREF inline Bit32u *loadRM32 ();
STH_PREF inline Bit8u  *loadReg8 ();
STH_PREF inline Bit16u *loadReg16();
STH_PREF inline Bit32u *loadReg32();

STH_PREF inline Bit16u *loadSReg();

STH_PREF inline Bit32u *loadCReg();
STH_PREF inline Bit32u *loadDReg();

STH_PREF inline Bit8u  *loadImm8 ();
STH_PREF inline Bit16u *loadImm16();
STH_PREF inline Bit32u *loadImm32();

STH_PREF inline Bit16u *loadSib16();
STH_PREF inline Bit32u *loadSib32();

STH_PREF inline Bit8u  *load_es_di_8 ();
STH_PREF inline Bit16u *load_es_di_16();
STH_PREF inline Bit32u *load_es_di_32();

STH_PREF inline Bit8u  *load_ds_si_8 ();
STH_PREF inline Bit16u *load_ds_si_16();
STH_PREF inline Bit32u *load_ds_si_32();

STH_PREF inline Bit8u  *loadRel8 ();
STH_PREF inline Bit16u *loadRel16();
STH_PREF inline Bit32u *loadRel32();

STH_PREF inline Bit8u  *loadMX8 ();
STH_PREF inline Bit16u *loadMX16();
STH_PREF inline Bit32u *loadMX32();


STH_PREF inline Bit16u *loadPtr16();
STH_PREF inline Bit32u *loadPtr32();

STH_PREF inline Bit8u  *CodeInc8 ();
STH_PREF inline Bit16u *CodeInc16();
STH_PREF inline Bit32u *CodeInc32();

STH_PREF inline void i_gonext();
STH_PREF inline void i_go();
STH_PREF inline void i_gonext_pm();
STH_PREF inline void i_go_pm();

// context
Bit8u modRM;
Bit8u sib;
char  amode,	// addres mode
      osize;	// operan size flag
      
char 	  segment;   // segment prefix
char 	  pref_rep;
char 	  pref_lock;
ModRMInfo modRMInfo;
Bit32s 	  disp16;
Bit32s    disp32;

STH_PREF inline void exception(int exc)
{
    longjmp(STH_THIS exc_buf,exc);
}

# ifndef JUST_DEP
#  include "nhandlers.h"
# endif

 private:

  Bit8u *CodeSeg;
  Bit8u *CodePtr;
  jmp_buf exc_buf;
  
STH_PREF   inline void sync_eip(){
    STH_THIS EIP=(STH_THIS CodePtr - STH_THIS CodeSeg) & 0xffff;
  }
STH_PREF   inline void sync_code(){
    STH_THIS CodeSeg=(Bit8u *)(STH_THIS mem+(_CS<<4));
    STH_THIS CodePtr=STH_THIS CodeSeg+(STH_THIS EIP & 0xffff);
  }

STH_PREF   inline void sync_code_ptr(){
    STH_THIS CodePtr=STH_THIS CodeSeg+(STH_THIS EIP&0xffff);
  }
};

#include "flags.h"

#endif
