/*
*  This file is part of Emma Virtual Machine
*  Copyright (C) 2003 Vitja Makarov <no1sm@narod.ru>
*  
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "cpu.h"

const char parity_matrix[]={
1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
};//parity matrix

const Bit32u val_masks[]={
        0x0,//0	
       0xff,//1
     0xffff,//2
   0xffffff,//3
 0xffffffff,//4
};
const Bit32u sign_bits[]={
	0x0,//0
       0x80,//1
     0x8000,//2
   0x800000,//3
 0x80000000,//4
};

#if 0
//add
void CPU::set_cf_like_add(Bit32u o1,Bit32u o2,Bit32u &res,int){
set_bit(eflags,CF,res<o1||res<o2);
}

void CPU::set_of_like_add(Bit32u o1,Bit32u o2,Bit32u &res,int size){
Bit32u s=sign_bit(size);
set_bit(eflags,OF,((o1&s)==(o2&s)) && ((res&s)!=(o2&s)));
}
void CPU::set_af_like_add(Bit32u o1,Bit32u o2,Bit32u &res,int){
set_bit(eflags,AF,((o1^o2) ^res) & 0x10  );
}
//sub
void CPU::set_cf_like_sub(Bit32u o1,Bit32u o2,Bit32u &res,int){
set_bit(eflags,CF,o1<o2);
}
void CPU::set_of_like_sub(Bit32u o1,Bit32u o2,Bit32u &res,int size){
Bit32u s=sign_bit(size);
set_bit(eflags,OF, //((o1&s)!=(o2&s)) && ((res&s)==(o2&s));
( (o1 ^ o2) & (o1 ^ res) & s) );
}
void CPU::set_af_like_sub(Bit32u o1,Bit32u o2,Bit32u &res,int){
set_bit(eflags,AF, ( ((o1^o2) ^res) & 0x10 ) );
}


void CPU::set_flags_add(Bit32u o1,Bit32u o2,Bit32u &res,int s)
{
res&=val_mask(s);
set_cf_like_add(o1,o2,res,s);
set_of_like_add(o1,o2,res,s);
set_af_like_add(o1,o2,res,s);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
return;
}
void CPU::set_flags_inc(Bit32u o1,Bit32u o2,Bit32u &res,int s)
{
res&=val_mask(s);
set_of_like_add(o1,o2,res,s);
set_af_like_add(o1,o2,res,s);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
return;
}

void CPU::set_flags_adc(Bit32u o1,Bit32u o2,Bit32u &res,int s)
{
int prev_cf;
res&=val_mask(s);
prev_cf=get_bit(eflags,CF);
set_cf_like_add(o1,o2,res,s);//!!!
//else one case
set_bit(eflags,CF,get_bit(eflags,CF) || (prev_cf&(o2==res)));
set_of_like_add(o1,o2,res,s);
set_af_like_add(o1,o2,res,s);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
return;
}

void CPU::set_flags_sub(Bit32u o1,Bit32u o2,Bit32u &res,int s)
{
res&=val_mask(s);
set_cf_like_sub(o1,o2,res,s);
set_of_like_sub(o1,o2,res,s);
set_af_like_sub(o1,o2,res,s);
set_szp(res,s);
return;
}

void CPU::set_flags_sbb(Bit32u o1,Bit32u o2,Bit32u &res,int s)
{
int prev_cf=get_bit(eflags,CF);
res&=val_mask(s);
set_cf_like_sub(o1,o2,res,s);
//else one case
set_bit(eflags,CF,get_bit(eflags,CF)||(prev_cf&(o2==o1)));
set_of_like_sub(o1,o2,res,s);
set_af_like_sub(o1,o2,res,s);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
return;
}
void CPU::set_flags_dec(Bit32u o1,Bit32u o2,Bit32u &res,int s)
{
res&=val_mask(s);
set_of_like_sub(o1,o2,res,s);
set_af_like_sub(o1,o2,res,s);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
return;
}

void CPU::set_flags_logical(Bit32u o1,Bit32u o2,Bit32u &res,int s){
//commands like OR,XOR,AND uses same flags-set scheme
res&=val_mask(s);
clc();
clo();
cla();
set_sf(res,s);
set_zf(res,s);
set_pf(res,s);
return;
}

void CPU::set_flags_shl(Bit32u o1,Bit32u o2,Bit32u &res,int s)
{
res&=val_mask(s);
o2&=0x1f;
    if(o2==0)
     return;
set_bit(eflags,CF,(o1>>(s*8-o2))&0x1);
  if(o2==1)
    set_bit(eflags,OF, ((res&sign_bit(s))?1:0)^get_bit(eflags,CF) );
set_bit(eflags,AF);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
}

void CPU::set_flags_shr(Bit32u o1,Bit32u o2,Bit32u &res,int s)
{
res&=val_mask(s);
  if(o2==0)
     return;
set_bit(eflags,CF,(o1>>(o2-1))&0x1);
  if(o2==1)
    set_bit(eflags,OF,(res&sign_bit(s))&&0x1);
    //((res&sign_bit(s))?1:0)^FLAGS.CF;
set_bit(eflags,AF,1);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
}

void CPU::set_flags_ror(Bit32u o1,Bit32u o2,Bit32u &res,int s)
{
res&=val_mask(s);
set_bit(eflags,CF,(o1>>(o2-1))&0x1);
//  if(o2==0)	this case release in the same function
//     return;
  if(o2==1)
    set_bit(eflags,OF,((res&sign_bit(s))^(o1>>(o2-1)))?1:0);
return;
}

void CPU::set_flags_rcl(Bit32u o1,Bit32u count,Bit32u &res,int s)
{
res&=val_mask(s);
set_bit(eflags,CF,(o1>>(s*8-count))&0x1);
  if(count==1)
    set_bit(eflags,OF,((res^o1)&sign_bit(s))?1:0 );
return;
}
#endif
