/*
*  This file is part of Emma Virtual Machine
*  Copyright (C) 2003 Vitja Makarov <no1sm@narod.ru>
*  
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _FLAGS_H_
#define _FLAGS_H_

/* flags */
#define CF  (1 <<  0)
#define PF  (1 <<  2)
#define AF  (1 <<  4)
#define ZF  (1 <<  6)
#define SF  (1 <<  7)
#define TF  (1 <<  8)
#define IF  (1 <<  9)
#define DF  (1 << 10)
#define OF  (1 << 11)
#define NT  (1 << 14)
#define RF  (1 << 16)
#define VM  (1 << 17)
#define AC  (1 << 18)
#define VIF VIF_MASK
#define VIP VIP_MASK
#define ID  ID_MASK

#ifndef IOPL_MASK
#define IOPL_MASK  (3 << 12)
#endif

extern const char parity_matrix[];
extern const Bit32u sign_bits[];
extern const Bit32u val_masks[];
#if 0
//size should be given in bytes
inline Bit32u sign_bit(int size){
return sign_bits[size];
}
inline Bit32u val_mask(int size){
return val_masks[size];
}
#else 
#endif
# define sign_bit(s) sign_bits[s]
# define val_mask(s) val_masks[s]

#if 1
# define  cls(){ clear_bit(__eflags,SF);}
# define  clz(){ clear_bit(__eflags,ZF);}
# define  clo(){ clear_bit(__eflags,OF);}
# define  clc(){ clear_bit(__eflags,CF);}
# define  cla(){ clear_bit(__eflags,AF);}
# define  clp(){ clear_bit(__eflags,PF);}
# define  cld(){ clear_bit(__eflags,DF);}
# define  cli(){ clear_bit(__eflags,IF);}
# define  sts(){ set_bit(__eflags,SF);}
# define  stz(){ set_bit(__eflags,ZF);}
# define  sto(){ set_bit(__eflags,OF);}
# define  stc(){ set_bit(__eflags,CF);}
# define  sta(){ set_bit(__eflags,AF);}
# define  stp(){ set_bit(__eflags,PF);}
# define  std(){ set_bit(__eflags,DF);}
# define  sti(){ set_bit(__eflags,IF);}
#endif

# if !defined (DEF_FLAGS)
//add
inline void CPU::set_cf_like_add(Bit32u o1,Bit32u o2,Bit32u res,int){
sc_bit(__eflags,CF,res<o1||res<o2);
}

inline void CPU::set_of_like_add(Bit32u o1,Bit32u o2,Bit32u res,int size){
Bit32u s=sign_bit(size);
sc_bit(__eflags,OF,((o1&s)==(o2&s)) && ((res&s)!=(o2&s)));
}
inline void CPU::set_af_like_add(Bit32u o1,Bit32u o2,Bit32u res,int){
sc_bit(__eflags,AF,((o1^o2) ^res) & 0x10  );
}
//sub
inline void CPU::set_cf_like_sub(Bit32u o1,Bit32u o2,Bit32u res,int){
sc_bit(__eflags,CF,o1<o2);
}
inline void CPU::set_of_like_sub(Bit32u o1,Bit32u o2,Bit32u res,int size){
Bit32u s=sign_bit(size);
sc_bit(__eflags,OF, //((o1&s)!=(o2&s)) && ((res&s)==(o2&s));
( (o1 ^ o2) & (o1 ^ res) & s) );
}
inline void CPU::set_af_like_sub(Bit32u o1,Bit32u o2,Bit32u res,int){
sc_bit(__eflags,AF, ( ((o1^o2) ^res) & 0x10 ) );
}

inline void CPU::set_flags_add(Bit32u o1,Bit32u o2,Bit32u res,int s)
{
set_cf_like_add(o1,o2,res,s);
set_of_like_add(o1,o2,res,s);
set_af_like_add(o1,o2,res,s);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
return;
}

inline void CPU::set_flags_inc(Bit32u o1,Bit32u o2,Bit32u res,int s)
{
set_of_like_add(o1,o2,res,s);
set_af_like_add(o1,o2,res,s);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
return;
}

inline void CPU::set_flags_adc(Bit32u o1,Bit32u o2,Bit32u res,int s)
{
int prev_cf;

prev_cf=get_bit(__eflags,CF);
set_cf_like_add(o1,o2,res,s);//!!!
//else one case
sc_bit(__eflags,CF,get_bit(__eflags,CF) || (prev_cf&(o2==res)));
set_of_like_add(o1,o2,res,s);
set_af_like_add(o1,o2,res,s);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
return;
}

inline void CPU::set_flags_sub(Bit32u o1,Bit32u o2,Bit32u res,int s)
{
set_cf_like_sub(o1,o2,res,s);
set_of_like_sub(o1,o2,res,s);
set_af_like_sub(o1,o2,res,s);
set_szp(res,s);
return;
}

inline void CPU::set_flags_sbb(Bit32u o1,Bit32u o2,Bit32u res,int s)
{
int prev_cf=get_bit(__eflags,CF);

set_cf_like_sub(o1,o2,res,s);
//else one case

sc_bit(__eflags,CF,get_bit(__eflags,CF)||(prev_cf&(o2==o1)));

set_of_like_sub(o1,o2,res,s);
set_af_like_sub(o1,o2,res,s);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
return;
}
inline void CPU::set_flags_dec(Bit32u o1,Bit32u o2,Bit32u res,int s)
{
set_of_like_sub(o1,o2,res,s);
set_af_like_sub(o1,o2,res,s);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
return;
}

inline void CPU::set_flags_logical(Bit32u o1,Bit32u o2,Bit32u res,int s){
//commands like OR,XOR,AND uses same flags-set scheme
clc();
clo();
cla();
set_sf(res,s);
set_zf(res,s);
set_pf(res,s);
return;
}

inline void CPU::set_flags_shl(Bit32u o1,Bit32u o2,Bit32u res,int s)
{
sc_bit(__eflags,CF,(o1>>(s*8-o2))&0x1);
  if(o2==1)
    sc_bit(__eflags,OF, ((res&sign_bit(s))?1:0)^get_bit(__eflags,CF) );
set_bit(__eflags,AF);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
}

inline void CPU::set_flags_shr(Bit32u o1,Bit32u o2,Bit32u res,int s)
{
sc_bit(__eflags,CF,(o1>>(o2-1))&0x1);
  if(o2==1)
    sc_bit(__eflags,OF,(res&sign_bit(s))&&0x1);
    //((res&sign_bit(s))?1:0)^FLAGS.CF;
set_bit(__eflags,AF);
set_pf(res,s);
set_zf(res,s);
set_sf(res,s);
}

inline void CPU::set_flags_ror(Bit32u o1,Bit32u o2,Bit32u res,int s)
{
sc_bit(__eflags,CF,(o1>>(o2-1))&0x1);
  if(o2==1)
    sc_bit(__eflags,OF,((res&sign_bit(s))^(o1>>(o2-1)))?1:0);
return;
}

inline void CPU::set_flags_rcl(Bit32u o1,Bit32u count,Bit32u res,int s)
{
sc_bit(__eflags,CF,(o1>>(s*8-count))&0x1);
  if(count==1)
    sc_bit(__eflags,OF,((res^o1)&sign_bit(s))?1:0 );
return;
}
# else

#define set_cf_like_add(o1,o2,res,___) \
    sc_bit (__eflags, CF, (res) < (o1) || (res) < (o2));


#define set_of_like_add(o1,o2,res,size)	\
  sc_bit (__eflags, OF, (( (o1) & sign_bit (size)) == \
			    ( (o2) & sign_bit (size))) \
	&& (( (res) & sign_bit (size)) != ( (o2) & sign_bit (size))));

#define set_af_like_add(o1,o2,res,___) \
    sc_bit (__eflags, AF, (( (o1) ^ (o2)) ^ res) & 0x10);


#define set_cf_like_sub(o1,o2,res,___) \
    sc_bit(__eflags,CF,(o1)<(o2));

#define set_of_like_sub(o1,o2,res,size)	\
    sc_bit(__eflags,OF,			\
	( ((o1) ^ (o2)) & ((o1) ^ (res)) & sign_bit(size)) );
	
#define set_af_like_sub(o1,o2,res,___)	\
    sc_bit(__eflags,AF, ( (( (o1) ^ (o2) ) ^ (res) ) & 0x10 ) );

#define set_flags_add(o1,o2,res,s)\
{	\
    set_cf_like_add(o1,o2,res,s);	\
    set_of_like_add(o1,o2,res,s);	\
    set_af_like_add(o1,o2,res,s);	\
    set_pf(res,s);			\
    set_zf(res,s);			\
    set_sf(res,s);			\
}

#define set_flags_inc(o1,o2,res,s)	\
{					\
set_of_like_add(o1,o2,res,s);		\
set_af_like_add(o1,o2,res,s);		\
set_pf(res,s);				\
set_zf(res,s);				\
set_sf(res,s);				\
}

#define set_flags_adc( o1, o2, res, s) 	\
{					\
int prev_cf;				\
					\
prev_cf=get_bit(__eflags,CF);		\
set_cf_like_add(o1,o2,res,s);		\
sc_bit(__eflags,CF,get_bit(__eflags,CF) || (prev_cf&((o2)==(res)))); \
set_of_like_add(o1,o2,res,s);			\
set_af_like_add(o1,o2,res,s);			\
set_pf(res,s);					\
set_zf(res,s);					\
set_sf(res,s);					\
}

#define set_flags_sub(o1,o2,res,s)	\
{					\
set_cf_like_sub(o1,o2,res,s);		\
set_of_like_sub(o1,o2,res,s);		\
set_af_like_sub(o1,o2,res,s);		\
set_szp(res,s);				\
}

#define set_flags_sbb(o1,o2,res,s)	\
{					\
int prev_cf=get_bit(__eflags,CF);	\
set_cf_like_sub(o1,o2,res,s);		\
					\
sc_bit(__eflags,CF,get_bit(__eflags,CF)||(prev_cf&( (o2) == (o1) )));\
				\
set_of_like_sub(o1,o2,res,s);	\
set_af_like_sub(o1,o2,res,s);	\
set_pf(res,s);			\
set_zf(res,s);			\
set_sf(res,s);			\
}

#define set_flags_dec(o1,o2,res,s)	\
{					\
set_of_like_sub(o1,o2,res,s);		\
set_af_like_sub(o1,o2,res,s);		\
set_pf(res,s);				\
set_zf(res,s);				\
set_sf(res,s);				\
}

#define set_flags_logical(o1,o2,res,s){	\
clc();	\
clo();	\
cla();	\
set_sf(res,s);	\
set_zf(res,s);	\
set_pf(res,s);	\
}

#define set_flags_shl(o1,o2,res,s)	\
{					\
sc_bit(__eflags,CF,( (o1) >>( (s)*8-o2))&0x1);	\
  if( (o2) == 1)				\
    sc_bit(__eflags,OF, (( (res) &sign_bit(s))?1:0)^get_bit(__eflags,CF) ); \
set_bit(__eflags,AF);	\
set_pf(res,s);		\
set_zf(res,s);		\
set_sf(res,s);		\
}

#define set_flags_shr(o1,o2,res,s)	\
{					\
sc_bit(__eflags,CF,( (o1) >>(o2-1))&0x1);	\
  if( (o2) ==1)					\
    sc_bit(__eflags,OF,(res&sign_bit(s))&&0x1);	\
set_bit(__eflags,AF);		\
set_pf(res,s);			\
set_zf(res,s);			\
set_sf(res,s);			\
}

#define set_flags_ror(o1,o2,res,s)		\
{						\
sc_bit(__eflags,CF,( (o1) >>(o2-1))&0x1); 	\
  if((o2)==1)					\
    sc_bit(__eflags,OF,(( (res) &sign_bit(s))^( (o1) >>(o2-1)))?1:0);	\
}

#define set_flags_rcl(o1,count,res,s)	\
{					\
sc_bit(__eflags,CF,( (o1) >>((s)*8-(count)))&0x1);	\
  if((count) == 1)					\
    sc_bit(__eflags,OF,(( (res) ^ (o1) )&sign_bit(s))?1:0 );	\
}
# endif

#endif /* flags.h */

