#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <unistd.h>

#include "dbg-txt.h"

#ifdef USE_READLINE
# include <readline/readline.h>
# include <readline/history.h>
extern HIST_ENTRY **history_list ();
#endif

#ifndef whitespace
#define whitespace(c) (((c) == ' ') || ((c) == '\t') || ((c) == '\n'))
#endif



#define CMD_QUIT 0x1
#define CMD_NEXT 0x4
#define CMD_RUN  0x5
#define CMD_SUSPEND 0x6

extern volatile int in_debug;


struct FLGS{
char * n /* name */;
Bit32u m /* mask */;
};

static struct FLGS FLGS[] = {
    {"CF", 0x01},
    {"ZF", 0x40},
    {"SF", 0x80},
    {"OF", 0x800},
    {"PF", 0x4},
    {"AF", 0x10},
    {"IF", 0x200},
    {"DF", 0x400},
    {NULL, 0}
};

//extern int no_timer;

CPU *dbgCPU;
disasm *dhost;
char *getline (char * prompt)
{
#ifdef USE_READLINE
    return readline(prompt);
#else
#define MAX_STRLEN 200
    char *str;
	if(prompt)
	    printf("%s",prompt);
	    
    str = (char *)malloc(MAX_STRLEN);
    str = fgets( str, MAX_STRLEN, stdin);
    return str;
#endif
}

#if 0
int cmd_tm(char *){
    no_timer ^= 1;
    return 0;
}
#endif

DebugCMD cmds[] = {
  {"quit", cmd_quit, "Quit.."},
  {"q", cmd_quit, ""},
  {"exit", cmd_quit, ""},
  {"help", cmd_help, "Help screen"},
  {"next", cmd_next, "Step next"},
  {"n", cmd_next, ""},
  {"run", cmd_run, "Run"},
  {"r", cmd_run, ""},
  {"c", cmd_run, "Continue"},
  {"b", cmd_bp, "breakpoint [SEG:OFF] | [OFF] | (nil) for current point"},
  {"bs",cmd_bs, "break on string: bs 0xXX 0xXX 0xXX..."},
  {"d", cmd_dbp, "Disable breakpoint <bp's number>"},
  {"e", cmd_ebp, "Enable  breakpoint <bp's number>"},
  {"dis", cmd_dis, "Disassemble CS:eIP"},
# if defined (MIPS_COUNTER)  
  {"mips", cmd_mips, "Print current MIPS"},
# endif  
  {"regs", cmd_regs, "Print cpu registers"},
  {"shell", cmd_shell, "run shell command"},
  {"make", cmd_make, "run make"},
  {"warranty", cmd_warranty, "shows warranty"},
  {"mkr", cmd_mkr, "Make and restart emma"},
  {"mem", cmd_mem, "Memory dump, mem <[abs addres]|[seg:off]> [lines]"},
  {"log", cmd_log, "On/Off logging mode"},
#if defined (DBG_EXPR)  
  {"p",   cmd_pdec, "Print value or expression (dec)"},
  {"p/",  cmd_phex,  "(hex)"},
  {"p/o", cmd_poct, "(oct)"},
  {"p/b", cmd_pbin, "(bin)"},
#endif  
  {"backtrace", cmd_backtrace, "Enough stupid backtrace"},

  {"s", cmd_suspend, "Kill self and go into gdb if any"},
  {"z", cmd_suspend, "Kill self and go into gdb if any"},
  {"soe", cmd_soe, "Stop on exception on/off"},
//  {"tm", cmd_tm, ""},


//  {"regs32", cmd_no, ""},
//  {"regs", cmd_no, ""},
//  {"mem", cmd_no, "mem addr"},
//  {"setmem", cmd_no, "setmem addr char"},
  {(char *) NULL, (FuncHandler *) NULL, (char *) NULL},
};

static int __first = 1;

#define MAX_BPS 100

#define BP_IP   1
#define BP_CODE 2
#define BP_TMP  3

struct BP
{
  Bit16u seg;
  Bit32u off;
  int on;
  int i_len;
  int type;
  char *i_data;
};

#define BT_CALL  0x10
#define BT_LCALL 0x11
#define BT_RET   0x20
#define BT_LRET  0x21
#define BT_JMP   0x40
#define BT_LJMP  0x41

struct BackTraceInfo{
Bit16u seg;
Bit32u off;
Bit16u oseg;
Bit32u ooff;
int type;
};

static struct BP bps[MAX_BPS];
static int bps_num = 0;
static FILE * out_fp=NULL;
static BackTraceInfo *bt_info=NULL;
static int bt_info_len=0;
static int stop_onexc=1;

static void dumpRegs(FILE *fp);



/* backtrace stack */

inline int backtr_init()
{
    if(!bt_info)
    {
        bt_info_len=0;
	bt_info = (BackTraceInfo *) malloc(sizeof(struct BackTraceInfo));
    }
    else
    {
	if(bt_info_len<0)
	    return 1;
    
	bt_info = (BackTraceInfo *) realloc(bt_info,
		 sizeof(struct BackTraceInfo) * (bt_info_len+1) );
    }
return 0;    
}

static void backtr_push(int type, Bit16u oldSeg, Bit32u oldOff)
{
     if(backtr_init())
        return ;
    
    bt_info[bt_info_len].seg  = dbgCPU->SRegs[_cs_];
    bt_info[bt_info_len].off  = dbgCPU->EIP;
    bt_info[bt_info_len].oseg  = oldSeg;
    bt_info[bt_info_len].ooff  = oldOff;
    bt_info[bt_info_len].type = type;
    
    bt_info_len++;
}

static void backtr_pop()
{
    if(!bt_info)
      return;
      for(;;)
      {
	if(bt_info_len)
	{
	    if(bt_info[--bt_info_len].type & BT_JMP)
		/* got jump -> skip */;
	    else
		return;
	}
	else
	  return;
      }
}


void
SigInt (int)
{
  in_debug = 1;
}

void
about ()
{
  fprintf (out_fp,"Emma %s, Copyright (C) 2003 Vitja Makarov\n"
	  "Emma comes with ABSOLUTELY NO WARRANTY; for details type `warranty'.\n"
	  "This is free software, and you are welcome to redistribute it\n"
	  "under certain conditions.\n\n", 
#ifdef EMMA_VERSION	  
	  EMMA_VERSION
#else
	  ""
#endif	  
	  );
}

static int log_level=0;
static int bt_full=0;

#define LOG_DIS   1
#define LOG_REGS  2
#define LOG_NBIOS 4
#define LOG_NDIS  8 /* opcode and etc */

void
debug_go (CPU * cpu,int i_ret)
{
  static char *temp = NULL, *prompt= "(emma-dbg) ", *last = NULL;
  static int done=0;
  int ret;
  static int in_init=0;
  static FILE *init_fp;
  static  Bit16u oldCS=0;
  static  Bit32u oldIP=0;
  static  int bt_susp=0;

  if (__first){
      dbgCPU=cpu;
      out_fp=stdout;
      
      signal (SIGINT, SigInt);
      about ();
      __first = 0;
      dhost = new disasm();
      dhost-> buf = dbgCPU->mem;
      dhost-> buflen = 1024*1024;
      dhost-> bufpos = 0;
      
       if((init_fp=fopen(".emma-dbg","r"))){
        in_init=1;
	    fprintf(out_fp,"Using init script\n");
	}
    }
// do backtrace
    if(bt_susp)
     {
	if(bt_susp&BT_RET && !bt_full)
	    backtr_pop();
	else
	    backtr_push(bt_susp,oldCS,oldIP);
        bt_susp=0;
     }
// prefix less
int  off=-1;
//int  off_len,a32,o32;

next_again:
    off++;
         
    switch(*(Bit8u *)dbgCPU->memptr(dbgCPU->SRegs[_cs_],dbgCPU->EIP+off))
    {
    case 0xff: /* GRP 5 */
	switch(	(*(Bit8u *)dbgCPU->memptr(dbgCPU->SRegs[_cs_]
					,dbgCPU->EIP + (++off)))
	    & 0x38
		)
	{
	case 0x10: /* call  rm */
		bt_susp=BT_CALL;
		
		break;
	case 0x18: /* lcall rm */
		bt_susp=BT_LCALL;
		break;
	case 0x20: /* ljmp  rm */
		bt_susp=BT_JMP;
		break;
	}
	
    break;
    case 0x9a: /* lcall */
    		bt_susp=BT_LCALL;
		break;
    case 0xe8: /* call  */
    		bt_susp=BT_CALL;
		break;
    case 0xea: /* ljmp  */
    		bt_susp=BT_LJMP;
		break;
    break;
    case 0xc2:/* ret	 */
    case 0xc3:/* ret %d	 */
	bt_susp=BT_RET;
	break;
    case 0xca:/* lret	 */
    case 0xcb:/* lret %d */
	bt_susp=BT_LRET;
	break;
    case 0x66:
    case 0x67:
    case 0xf0:
	goto next_again;
    
    }
    
    
  for (int i = 0; i < bps_num ; i++)
    {
     if(bps[i].on)
    	switch(bps[i].type)
	{
  
	case BP_IP:
	if (dbgCPU->SRegs[_cs_] == bps[i].seg && dbgCPU->EIP == bps[i].off)
          {
        	in_debug = 1;
		fprintf(out_fp,"\nBreakpoint %d, %04x:%04lx\n\n", i + 1, bps[i].seg,
			bps[i].off);
		    if(bps[i].on & BP_TMP)
			bps[i].on = 0;	
		break;
	  }
	break;
	case BP_CODE:
	  if(!memcmp(dbgCPU->memptr(dbgCPU->SRegs[_cs_],dbgCPU->EIP),bps[i].i_data,bps[i].i_len))
	    {
		fprintf(out_fp,"\nBreakpoint %d...\n\n",i+1);
		in_debug=1;
	    }
	
	break;
	}
    }
    
    if(i_ret && stop_onexc)

    switch(i_ret)
    {
	case 0:
//	default: /* Ok */
	    break;
	break;
	default:
//	case EXC_UD:
	    out_fp=stdout;
	    cmd_disone(oldCS,oldIP);
	    fprintf(out_fp,"\nUndefined instruction.\t^^^^^^^^....stoping...%d\n",i_ret);
	    in_debug=1;
    	    break;
    }

  oldIP=cpu->EIP;
  oldCS=cpu->SRegs[_cs_];


  if (!in_debug)
    {
	out_fp=stderr;
	 if(log_level & LOG_NBIOS && dbgCPU->SRegs[_cs_]==0xf000)
	    return;
	    if(log_level&LOG_REGS){
		    dumpRegs(out_fp);
        	    fprintf(out_fp,"\n");
	    }
	    if(log_level & LOG_DIS)
    		cmd_disone (dbgCPU->SRegs[_cs_],dbgCPU->EIP, log_level & LOG_NDIS);
      return;
    }
  out_fp=stdout;
  dbgCPU = cpu;
// print current instruction
  if(!i_ret)
    {
      cmd_regs ("");
      fprintf(out_fp,"\n");
      cmd_disone (dbgCPU->SRegs[_cs_],dbgCPU->EIP);
    }

  while (!done)
    {
	if(in_init){
	    temp=(char *)malloc(100);
	    if(!fgets(temp,100,init_fp)){
		fclose(init_fp);
		in_init=0;
		free(temp);
		}else	  
		if(temp[0]=='#'){
		    free(temp);
		    continue;
		    }
	
	}
	    
	if(!in_init)
	  temp = getline ( prompt );

      /* Test for EOF. */
      if (!temp)
	return;

      if (!temp[0] || temp[0]=='\n' && last)
	{
	  free (temp);
	  temp = last;
	}
      else if (last)
	free (last);

      /* If there is anything on the line, print it and remember it. */
      if (*temp)
	{  
#ifdef USE_READLINE	
	if(temp!=last && !in_init)
	     add_history (temp);
#endif	     
	  ret = execute_command (temp);
	 if(!in_init)
          last = temp;
	  
	  switch (ret)
	    {
	    case CMD_RUN:
	      in_debug = 0;
	      return;
	      break;
	    case CMD_NEXT:
	      return;
	      break;
	    case CMD_QUIT:
	      exit (0);
	      break;
	    case CMD_SUSPEND:
	      kill (0, SIGINT);
	      return;
	    }
	}
    }
//  exit (0);
  return;
}





int
execute_command (char *line)
{
  int i = 0;
  char *word;
  DebugCMD *com;
  if (line == NULL)
    return 0;

  while (line[i] && whitespace (line[i]))
    i++;
  word = line + i;

  while (line[i] && !whitespace (line[i]))
    i++;
  if (line[i])
    line[i++] = '\0';

  com = find_command (word);

  if (!com)
    {
      fprintf (out_fp, "Undefined command: %s\n", word);
      return -1;
    }

  while (whitespace (line[i]))
    i++;
  word = line + i;

  return com->func (word);
}

DebugCMD *
find_command (char *name)
{
  int i;
  for (i = 0; cmds[i].name; i++)
    if (strcmp (name, cmds[i].name) == 0)
      return (&cmds[i]);

  return ((DebugCMD *) NULL);
}

/*
char *nextarg(char *arg,int &i)
{
char *word;
  while (arg[i] && whitespace (arg[i]))
    i++;
  word = arg + i;

  while (arg[i] && !whitespace (arg[i]))
    i++;
  if (arg[i])
    arg[i++] = '\0';
return word;
}
*/
int
cmd_quit (char * /*arg */ )
{
//  exit (0);
  return CMD_QUIT;
}

int
cmd_help (char * /*arg */ )
{
  int i;
  for (i = 0; cmds[i].name; i++)
    fprintf (out_fp, "%s\t\t%s\n", cmds[i].name, cmds[i].doc);
  return 0;
}

int
cmd_no (char * /*arg */ )
{
  fprintf (out_fp, "Sorry...\n");
  return 0;
}

int
cmd_next (char *)
{
  return CMD_NEXT;
}

int
cmd_run (char *)
{
  return CMD_RUN;
}

int
cmd_disone(Bit16u seg, Bit32u off,int a)
{
  dhost->com.lip= off;
  dhost->com.ip = off;
  dhost->bufpos = seg * 0x10 + off;

 fprintf(out_fp,"%04x:%04lx\t%s\n",
	    seg, dhost->com.ip, dhost->dis (F_ALL & (~(F_IP | (a? F_DECODE : 0)) ) ));
 return 0;
}

int
cmd_dis (char *args)
{
  int num=0;
  int ret;
  Bit16u seg;
  Bit16u ip;

  ret = sscanf (args, "%x:%x %d", &seg, &ip, &num);
  if (ret == 2);
//    num = 8;
  else  if (ret < 2)
    {
      ret = sscanf (args, "%x %d", &ip,&num);
      seg = dbgCPU->SRegs[_cs_];
       if(ret<1)
          ip = dbgCPU->EIP;
    }


  dhost->com.lip = ip;
  dhost->com.ip = ip;
  dhost->bufpos = seg * 0x10 + ip;
 if(num>0)
   for (int i = 0; i < num; i++)
 fprintf(out_fp,"%04x:%04lx\t%s\n",
	    seg, dhost->com.ip, dhost->dis (F_ALL & (~F_IP)));
  else{
     for(;;){
         for (int i = 0; i < 20; i++)
	     fprintf(out_fp,"%04x:%04lx\t%s\n",
			seg, dhost->com.ip, dhost->dis (F_ALL & (~F_IP)));
	    fprintf(out_fp,"--- Type <return> to continue,"
	    " or q <return> to quit---");
	    fflush(out_fp);
		if(getchar()=='q')
		    return 0;
	    fprintf(out_fp,"\n");
	}
      }

//    dbgCPU->parse_next();

//printf("unu=%d\n",ret);
  return 0;
}

int
cmd_regs (char *)
{
  dumpRegs (out_fp);
  return 0;
}

int
cmd_bp (char *args)
{
  Bit32u seg;
  Bit32u off;
  int ret;
  if (bps_num > MAX_BPS)
    {
      fprintf (out_fp, "error: Max breakpoints number reached\n");
      return 0;
    }
  ret = sscanf (args, "%lx:%lx", &seg, &off);

  switch (ret)
    {
    case 2:
      break;
    case 1:
      off = seg;
      seg = dbgCPU->SRegs[_cs_];
      break;
    default:
      seg = dbgCPU->SRegs[_cs_];
      off = dbgCPU->EIP;
    }

  seg&=0xffff;
  
  bps[bps_num].seg = seg;
  bps[bps_num].off = off;
  bps[bps_num].on = 1;
  bps[bps_num].type = BP_IP;


  bps_num++;
  fprintf(out_fp,"Breakpoint %d at %04lx:%04lx, emma-VM CPU.\n", bps_num, seg, off);
  return 0;
}

int cmd_bs(char *a)
{
char *next;
char *str;
int len=0;

    str = (char *)malloc(100);
    next = a;
printf("BPS: ");
    for(;next;next = strchr(next,','),len ++)
    {
try_next:    
	if(*next==',' || isspace(*next))
	    {
		next++;
		goto try_next;
	    }
	if ( sscanf(next,"0x%x",&str[len])<1 )
	{
	    fprintf(out_fp,"Bad!!\n");
	    break;
	}
	fprintf(out_fp, "%02x ",str[len]);
    }
        if(len==0){
	    fprintf(out_fp,"Bad!!\n");
		free(str);
		return 0;
	    }


    str = (char *)realloc(str,len);
    fprintf(out_fp,"\n");

  bps[bps_num].i_len =   len;
  bps[bps_num].i_data =  str;
  bps[bps_num].on = 1;
  bps[bps_num].type = BP_CODE;
    bps_num++;
    return 0;
}

int
cmd_dbp (char *a)
{
  int bpn;
  if (sscanf (a, "%d", &bpn) <= 0)
    {
      fprintf (out_fp, "syntax: d [number]\n");
      return 0;
    }
  if (bps_num < bpn)
    {
      fprintf (out_fp, "No such breakpoint %d\n", bpn);
      return 0;
    }
  fprintf (out_fp, "Disabling breakpoint %d at %04x:%04lx\n", bpn,
	   bps[bpn - 1].seg, bps[bpn - 1].off);
  bps[bpn - 1].on = 0;
  return 0;
}

int
cmd_ebp (char *a)
{
  int bpn;
  if (sscanf (a, "%d", &bpn) <= 0)
    {
      fprintf (out_fp, "syntax: e [number]\n");
      return 0;
    }
  if (bps_num < bpn)
    {
      fprintf (out_fp, "No such breakpoint %d\n", bpn);
      return 0;
    }
  fprintf (out_fp, "Enabling breakpoint %d at %04x:%04x\n", bpn,
	   bps[bpn - 1].seg, bps[bpn - 1].off);
  bps[bpn - 1].on = 1;
  return 0;
}

#if defined (MIPS_COUNTER)
int
cmd_mips (char *a)
{
  extern double mips;
  fprintf(out_fp,"MIPS=%.04f\n", mips);
  return 0;
}
#endif

int
cmd_suspend (char *a)
{
  return CMD_SUSPEND;
}

int
cmd_shell (char *a)
{
  if (a[0])
    system (a);
  else
    system ("sh");
  return 0;
}


int
cmd_make (char *a)
{
  char *arg;
  if (a[0])
    arg = (char *) malloc (strlen (a) + 10);
  else
    arg = (char *) malloc (10);

  sprintf (arg, "make %s", a);
  cmd_shell (arg);;
  free (arg);
  return 0;
}

int
cmd_mkr (char *a)
{
  cmd_make (a);			// here should be fork and exec
  // so... doit
//    execlp("./emma","./emma");
  cmd_shell ("./emma");
  exit (1);
}

int
cmd_mem (char *a)
{
int ret;
Bit32u seg,off,lines;
char *memptr;

    if(strchr(a,':')){
        ret=sscanf(a,"%lx:%lx %ld",&seg,&off,&lines);
	if(ret<2){
  	 fprintf(out_fp,"Invalid arguments:\nsyntax: mem <[off]|[seg:off]>\n");
	    return 0;
        }else if(ret<3) 
	    lines=10;

    }else{
        ret=sscanf(a,"%lx %ld",&off,&lines);
	    if(ret<1){
	     fprintf(out_fp,"Invalid arguments:\nsyntax: mem <[off]|[seg:off]>\n");
	     return 0;
	     }else if(ret<2)
	      lines=10;
    seg=dbgCPU->SRegs[_ds_];
    }
fprintf(out_fp,"Dumping memory at address: %04x:%04x\n",seg,off);
memptr=(char *)dbgCPU->memptr(seg,off);
    for(int i=0;i<lines;i++){
	fprintf(out_fp,"%04x:%04x  ",seg,off+i*0x10);
	    for(int j=0;j<0x10;j++)
	     fprintf (out_fp,"%02x ",*memptr++);
	fprintf (out_fp,"\n");
    }
    
return 0;
}


int
cmd_log (char *a){
int level,ret;
    if((ret=sscanf(a,"%d",&level))>0){
	log_level=level;
	}
    else
	log_level=!log_level;
fprintf(out_fp,"Switching logging mode log level %d.\n",
	log_level);
return 0;
}

int
cmd_warranty (char *)
{
  fprintf (out_fp,"                            NO WARRANTY\n"
	  "\n"
	  "  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY\n"
	  "FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN\n"
	  "OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES\n"
	  "PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED\n"
	  "OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n"
	  "MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS\n"
	  "TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE\n"
	  "PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,\n"
	  "REPAIR OR CORRECTION.\n"
	  "\n"
	  "  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING\n"
	  "  WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR\n"
	  "  REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,\n"
	  "  INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING\n"
	  "  OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED\n"
	  "  TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY\n"
	  "  YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER\n"
	  "  PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE\n"
	  "  POSSIBILITY OF SUCH DAMAGES.\n");
  return 0;
}

int
cmd_copying (char *)
{
  return 0;
}

#if defined(DBG_EXPR)
char *expr_buf;
int  expr_bufpos;
extern int dbglex();
int dbg_ret;
int
cmd_print(char *arg)
{
int ret;
char *dd;
expr_buf=(char *)malloc(strlen(arg)+2);
sprintf(expr_buf,"%s\n",arg);
expr_bufpos=0;
        if((ret=dbglex()))
	    fprintf(out_fp,"%d errors found.\n",ret);
free(expr_buf);
return ret;
}

int cmd_pdec(char *a){
 if(cmd_print(a))
    return 0;
    fprintf(out_fp,"%d\n",dbg_ret);
    return 0;    
}

int cmd_phex(char *a){
 if(cmd_print(a))
    return 0;
    fprintf(out_fp,"0x%lx\n",(Bit32u)dbg_ret);
    return 0;
}

int cmd_poct(char *a){
 if(cmd_print(a))
    return 0;
    fprintf(out_fp,"0%lo\n",(Bit32u)dbg_ret);
    return 0;
}

int cmd_pbin(char *a){
 if(cmd_print(a))
    return 0;
    
    for(int i=0;i<32;i++)
	fprintf(out_fp,"%d",(dbg_ret&(1<<(32-i)))?1:0);
    fprintf(out_fp,"b\n");
    return 0;
}

#endif

int cmd_soe(char *a){
    stop_onexc^=1;
    fprintf(out_fp,"Stop on execution now is %s\n",stop_onexc ? "On" : "Off");
    return 0;
}

int cmd_backtrace(char *a){
    if(strstr(a,"full"))
    {
	bt_full^=1;
	return 0;
    }

    fprintf(out_fp,"Backtrace: \n");
    for(int i=0;i<bt_info_len;i++)
	{
	    fprintf(out_fp,"%d:\t [%04x:%04x] => [%04x:%04x]",i,
					 bt_info[i].oseg,
					bt_info[i].ooff,
					 bt_info[i].seg,
					bt_info[i].off);
		switch(bt_info[i].type)
		{
		case BT_LCALL:
		    fprintf(out_fp,"\tlcall\n");
	    	    break;
		case BT_CALL:
		    fprintf(out_fp,"\tcall\n");
		    break;
		case BT_LJMP:
    		    fprintf(out_fp,"\tljmp\n");
		    break;
		case BT_RET:
    		    fprintf(out_fp,"\tret\n");
		    break;
		case BT_LRET:
    		    fprintf(out_fp,"\tlret\n");
		    break;
		default:
		    fprintf(out_fp,"\n");
		}			
		
	}
    return 0;
}

static void dumpRegs(FILE * fp)
{
    // common registers
    for (int i = 0; i < 8; i++) {
	fprintf(fp, "%s:%08lX\t", _reg_names[i].reg[2], dbgCPU->Regs[i]);
	if (!((i + 1) % 3))
	    fprintf(fp, "\n");
    }
    fprintf(fp, "eip:%08lX\n", dbgCPU->EIP);
    // segment registers
    for (int i = 0; i < 6; i++) {
	fprintf(fp, "%s:%04X\t", _segreg_names[i], dbgCPU->SRegs[i]);
	if (!((i + 1) % 3))
	    fprintf(fp, "\n");
    }
    for (int i = 0; i < 8; i++)
	fprintf(fp, "%s=%d ", FLGS[i].n, (FLGS[i].m & dbgCPU->eflags) ? 1 : 0);
    fprintf(fp, "EFlags:%08lX\t", dbgCPU->eflags);
    fprintf(fp, "\n");
}

