%option noyywrap

%{
//
//	if you want debug flexed files use -L option with flex
//

extern char *expr_buf;
extern int  expr_bufpos;
#define ECHO

#define YY_INPUT(buf,result,max_size) 	\
    {					\
	int c=expr_buf[expr_bufpos++];		\
	result=  (c==0) \
		? YY_NULL : (buf[0] = c, 1); \
    }
%}

%{
#include <stdlib.h>
#include <string.h>

#include "cpu.h"

extern CPU *dbgCPU;

/* types */
/* operation */
#define I_OP  0x10
#define I_ADD 0x10
#define I_SUB 0x11
#define I_MUL 0x14
#define I_DIV 0x18
/* value or variable */
#define I_VAL  0
/* brackets */
#define I_BRO 0x20
#define I_BRC 0x21
#define I_MBRO 0x20
#define I_MBRC 0x21
/* off */
#define I_OFF 0x40

/* priority */
#define P_BR  10 /* 1+2*(23+23) - you'l do brackets first, but... */
#define P_MUL 1
#define P_DIV 1
#define P_ADD 2
#define P_SUB 2

#define P_VAL 10
#define P_OFF P_VAL

extern int dbg_ret;

struct Item{
  int type;
  int priority;
  int val;
};


static int sign=1;

static struct Item *items=NULL;
static int i_len=-1;
static int on_error=0;

static void init(void)
{
    if(items==NULL){
	items=(struct Item *)malloc(sizeof(struct Item));
	i_len=0;
	on_error=0;
	}
}

static void add(struct Item *l){
    if(items)
        items=(struct Item *)realloc(items,sizeof(struct Item)*(++i_len+1));
    else
	init();
    *(items+i_len)=*l;
//    *(items+i_len).o
}

inline struct Item * last()
{
    return (i_len>=0)? (items+i_len): NULL;
}

inline void disable(struct Item *i){
    i->type=    I_OFF;
    i->priority=P_OFF;
    i->val     =0;
}

inline void do_sign(){
struct Item i;
    if(last() && !(last()->type&I_OP)){
	i.type = (*yytext=='-')? I_SUB: I_ADD;
	i.priority = P_ADD;
	i.val=0;
	add(&i);
    }
    else
        sign= (*yytext=='-')? -1: 1;
}

inline void do_mul(){
struct Item i;
i.type = I_MUL;
i.priority = P_MUL;
i.val=0;
add(&i);
}

inline void do_div(){
struct Item i;
i.type = I_DIV;
i.priority = P_DIV;
i.val=0;
add(&i);
}

/* add number */
inline void do_stuff(int cur){
struct Item i;
i.val=cur*sign;
i.type=I_VAL;
i.priority=P_VAL;
add(&i);
sign=1;
}

inline void do_dec(){
  int cur=0;
  sscanf(yytext,"%d",&cur);
  do_stuff(cur);
}

inline void do_oct(){
  int cur=0;
  sscanf(yytext,"0%o",&cur);
  do_stuff(cur);
}

inline void do_hex(){
  int cur=0;
  sscanf(yytext,"0x%x",&cur);
  do_stuff(cur);
}

inline void do_bin(){
  int cur;
  int tmp,i;
  tmp=strlen(yytext);
      for(i=0,cur=0;tmp>=0;tmp--)
	   if(yytext[tmp]=='1' || yytext[tmp]=='0')
	    {
		cur+= ( yytext[tmp]== '1' )? 1<<i : 0;
		i++;
	    }
  do_stuff(cur);
}

inline void do_reg8(){
    int c;
    int i;
	for(i=0;i<8;i++){
	    if(strstr(_reg_names[i].reg[0],yytext))
		break;
	}    
    
    c=((i)<4 ? LOW(dbgCPU->Regs[i]): HIGH(dbgCPU->Regs[i-4]) );
    do_stuff(c);
}

inline void do_reg16(){
    int c;
    int i;
	for(i=0;i<8;i++){
	    if(strstr(_reg_names[i].reg[1],yytext))
		break;
	}    
    
    c= LOWORD(dbgCPU->Regs[i]);
    do_stuff(c);
}

inline void do_reg32(){
    int c;
    int i;
	for(i=0;i<8;i++){
	    if(strstr(_reg_names[i].reg[2],yytext))
		break;
	}    
    
    c= dbgCPU->Regs[i];
    do_stuff(c);
}

inline void do_sreg(){
    int c;
    int i;
	for(i=0;i<6;i++){
	    if(strstr(_segreg_names[i],yytext))
		break;
	}    
    
    c= dbgCPU->SRegs[i];
    do_stuff(c);
}

inline void not_done(){
    fprintf(stderr,"Can't do... not done yet.\n");
};

#define TEST_ft(i,cond) (i.type cond)
inline int binary_test(int start, int end, int n,
	struct Item ** left, struct Item ** right){
int i;
    if(n<=start || n>=end)
	return 0;
    for(i=(n-1);i>=start;i--)
    {
	*left=items+i;
	    if((*left)->type==I_VAL)
		break;
	if(i==start)
	    return 0;
    }
    
    for(i=(n+1);i<=end;i++)
    {
	*right=items+i;
	    if((*right)->type==I_VAL)
		break;
	if(i==end)
	    return 0;
    }
	
    return 1;	
}

static int 
do_calc(int start,int end)
{
 int i,j,p;
 int res=0;
 struct Item *left,*right;
    
  do{    
// look for maximum priority
  p=items[start].priority;    
  j=start;
    for(i=start;i<=end;i++)
     {
        if(items[i].priority<p)
	    p=items[i].priority,j=i;
     }
        if(p>=P_OFF)
	    break;
// and now do operation
	  if(!binary_test(start,end,j,&left,&right))
	    return ++on_error;

	switch(items[j].type){
	case I_MUL:
	res=left->val*right->val;
	break;
	case I_DIV:
	res=left->val/right->val;
	break;
	case I_ADD:
	res=left->val+right->val;
	break;
	case I_SUB:
	res=left->val-right->val;
	break;
	}
    
    disable(left);
    disable(right);
    items[j].val=res;
    items[j].priority=P_VAL;
    items[j].type=I_VAL;
    }while(p<P_OFF);
	
    items[start].val= (items[start].type==I_VAL)? items[start].val: 0;
    items[start].priority=P_VAL;
    items[start].type=I_VAL;
    
	for(i=(start+1);i<=end;i++)
	{
	    if(items[i].type==I_VAL)
		items[start].val+=items[i].val;
	    disable(&items[i]);
	}
    
return 0;
}

int brs=0;

inline void do_bropen(){
struct Item i;
i.val=0;
i.priority=P_BR;
i.type=I_BRO;
add(&i);
brs++;
}

inline void do_brclose(){
int ss;
// here we can do the stuff
    for(ss=i_len;ss>=0;ss--){
	    if(items[ss].type==I_BRO)
		break;
	    else if(ss==0)
		return;
	
	}
disable(&items[ss]);
do_calc(ss,i_len);
//
brs--;
}
static Bit16u segment;
inline void do_segment()
{
int ss;
    for(ss=i_len;ss>=0;ss--){
	    if(items[ss].type==I_VAL)
		break;
	    else if(ss==0)
		return;
	
	}
segment=items[ss].val;
}

inline void do_mbropen(){
struct Item i;
i.val=0;
i.priority=P_BR;
i.type=I_MBRO;
add(&i);
brs++;
}

inline void do_mbrclose(){
int ss;
int off;
// here we can do the stuff
    for(ss=i_len;ss>=0;ss--){
	    if(items[ss].type==I_MBRO)
		break;
	    else if(ss==0)
		return;
	
	}
disable(&items[ss]);
do_calc(ss,i_len);
off=items[ss].val;
items[ss].val=*(int *)(dbgCPU->mem+off+segment * 0x10);
//
brs--;
}

inline void kill_context(){
    free(items);
    items=NULL;
}

inline int do_end(){
    int i;
	if(!items)
	    return on_error;

    do_calc(0,i_len);
        
    for(i=0,dbg_ret=0;i<=i_len;i++)
     if(items[i].type==I_VAL)
	 dbg_ret+=items[i].val;

    kill_context();

	if(brs)
	    return ++on_error;
    return on_error;
}


%}

alpha	[A-Za-z]
dig	[0-9]
name	({alpha}|{dig}|\$)({alpha}|{dig}|\_|\.|\-|\/|\$)*
dec1	[1-9]{dig}*
dec2	[1-9]{dig}*
dec	{dec1}|{dec2}
dig_hex [0-9a-fA-F]
dig_oct [0-7]
dig_bin [0-1]
hex 	0x{dig_hex}*?
oct 	0{dig_oct}*?
bin  	({dig_bin}*?)"b"
/*number {dig}*/
sign    [+-]
mul     "*"
div     "/"

br_open  "("
br_close ")"

mbr_open  "["
mbr_close "]"

segment  ":"
end  [\n\0]

/* White space, context-separator... */
ws [ ]*

/* CPU registers */
reg8   [acdb][lh]
reg16  ([acdb]x)|"sp"|"bp"|"di"|"si"
reg32  [e]{reg16}
sreg   [edcsfg]s

%%

{dec}{ws}    do_dec();
{hex}{ws}    do_hex();
{oct}{ws}    do_oct();
{bin}{ws}    do_bin();
{sign}{ws}   do_sign();
{mul}{ws}    do_mul();
{div}{ws}    do_div();
{reg8}{ws}   do_reg8();
{reg16}{ws}  do_reg16();
{reg32}{ws}  do_reg32();
{sreg}{ws}   do_sreg();
{end}        return do_end();
{br_open}    do_bropen();
{br_close}   do_brclose();
{mbr_open}    do_mbropen();
{mbr_close}   do_mbrclose();
{segment}     do_segment();
%%


#ifndef yywrap
 // something
  int yywrap(void)    /* do this to avoid needing to do -lfl */
       {
            return(1);
       }
#endif
