/*
*  This file is part of Emma Virtual Machine
*  Copyright (C) 2003 Vitja Makarov ?<vitja@mail.ru>?
*  
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#define WARNING  fprintf(stderr,"%s:%d WARNING!! doit\n",__FILE__,__LINE__);

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "disasm.h"
#include "dis386.h"
#include "decodes.cpp"

#define S_ALL   3
#define S_MINUS 2
#define S_PLUS  1
#define S_NO    0

#if 0
// use of catch, try, throw constructions 
// make it's speed slow, disasm was specially made
// for emma-vm that wants some speed one some it's
// internal levels
#  define USE_CATCH
#endif


#define loadOperand(o,s){	\
    (o)=0;			\
    for(int i=0;i<(s);i++)	\
 	((Bit8u *)&(o))[i]=nextbyte();\
	}

#define decode_modrm(){ 	\
  if(com.amode){		\
     com=modrm32[com.modRM];	\
    if(com.need_sib){		\
     com=sibInfo[nextbyte()];	\
        if(!(com.modRM&0xc0) && com.base==0x5)	\
	    com.base=_zero_,com.disp=4;		\
    }						\
  }else						\
    com=modrm16[com.modRM];			\
    loadOperand(com.offset,com.disp);		\
}

#define loadmodrm() {		\
	com.modRM=nextbyte();	\
	decode_modrm();		\
	}



/* capital letters in template are macros */
char *
disasm::normname (char *src, int aflag, int dflag)
{
  static char name_buf[20];
  char *p;
  char *obufp;
  obufp=name_buf;
  
  for (p = src; *p; p++)
    {
      switch (*p)
	{
	default:
	  *obufp++ = *p;
	  break;
	case 'C':		/* For jcxz/jecxz */
	  if (aflag)
	    *obufp++ = 'e';
	  break;
//	case 'N':
//	  if ((prefixes & PREFIX_FWAIT) == 0)
//	    *obufp++ = 'n';
//	  break;
	case 'S':
	  /* operand size flag */
	  if (dflag)
	    *obufp++ = 'l';
	  else
	    *obufp++ = 'w';
	  break;
	case 'W':
	  /* operand size flag for cwtl, cbtw */
	  if (dflag)
	    *obufp++ = 'w';
	  else
	    *obufp++ = 'b';
	  break;
	}
    }
  *obufp = 0;
  return name_buf;
}


inline char *doSign(Bit32u in,int len,//len is lenght on in in bytes
   		      int fmt=S_ALL)  // as you know from 1 to 4
{				      // may be we should use long long
static char  ret[20]; 		      // for most freedom and DEC compatible
int s;
// bit, that show's sign of in is 8*len-1
s=(in>>(8*len-1))?1:0;
unsigned long r=(Bit32u)-1;
    if(s)
    {
    r>>=(8*(sizeof(in)-len));
    in=r-in+1;
    }
static char s1[2],s2[2];
s1[1]=s2[1]=0;
s1[0]=fmt&S_PLUS?'+':0;
s2[0]=fmt&S_MINUS?'-':0;
sprintf(ret,"%s0x%lx",s?s2:s1,in);
return ret;
}

disasm::disasm()
{
bitD=0;
sstate=Ok;
buf=NULL;
bufpos=buflen=0;
}

int disasm::normbuf(){
// it's virtual and should, return 1 if 
// it's allright
return 0;
}


char *disasm::do_out(int fmt)
{
static  char _ret[100];// I hope it's enought
int off;
int i;
disout=_ret;
off=com.lip-com.ip;

    if(off<=0)
	return NULL;
if(fmt&F_IP)  
 disout+=sprintf(disout,"%08lX  ",com.ip);
if(fmt&F_CODE){
        for(i=0;i<off;i++)
        {
         disout+=sprintf(disout,"%02X",com.data[i]);
        }
    disout+=sprintf(disout,"%*s",18-i*2,"");
    }
if(fmt&F_DECODE)
  if(!com.bad){
	if(com.pref_seg>=0)
          disout+=sprintf(disout,"%s ",_segreg_names[com.pref_seg]);
	if(com.pref_lock)
          disout+=sprintf(disout,"lock ");
	switch(com.pref_rep)
	{
	case 0:
	break;
	case 1:
          disout+=sprintf(disout,"rep ");
	  break;
	case 2:
	  disout+=sprintf(disout,"repne ");
	  break;
	}
      disout+=sprintf(disout,"%s ",normname(com.op->name,com.amode,com.osize));
           if(com.op->print)
		    (this->*com.op->print)();
   }else{
       switch(com.lip-com.ip){
       case 1:    
       disout+=sprintf(disout,"byte 0x%02x",com.data[0]);
       break;
       case 2:    
       disout+=sprintf(disout,"word 0x%02x%02x",com.data[0],com.data[1]);
       break;
       case 4:    
       disout+=sprintf(disout,"dword 0x%02x%02x%02x%02x",com.data[0],com.data[1],
    			com.data[2],com.data[3]);
       break;
       }
     }
return _ret;     
}

char *disasm::dis(int fmt){
parse_next();
return do_out(fmt);
}

#ifndef BIGCASE
int disasm::parse_next()
{
#ifdef USE_CATCH
try
#endif
{
makePrefixAOP();
   if(/*!com.op||*/!com.op->name||/*com.op->name[0]=='(' ||*/ com.bad){
#ifdef USE_CATCH
      throw 1;
#endif
	com.bad=1;
	return -6;
	}      
// decode if it's needed
       if(com.op->decode)
    	    (this->*com.op->decode)();
    }
#ifdef USE_CATCH
    catch(int bad){
	com.bad=bad;
        return -6;
    }
#endif
return 1;
}
#endif

//parse's prefixes and opcode
//return value is number of parsed bytes
//const cmd zeroCOM;
inline void disasm::makePrefixAOP()
{
int s;
com.ip=com.lip;
com.amode=com.osize=bitD;
com.pref_seg=noSEG;
com.pref_rep=com.pref_lock=com.bad=0;
//com.operand[0]=com.operand[1]=com.operand[2]=com.offset=com.segment=0;

// look for prefixes
notdone:    s=nextbyte();
	    switch(s)
	     {
	     case 0x66:
	     com.osize^=1;
	     goto notdone;
	     case 0x67:
	     com.amode^=1;
	     goto notdone;
	     case 0x26:
  	     com.pref_seg=_es_;
	     goto notdone;
	     case 0x2e:
	     com.pref_seg=_cs_;
	     goto notdone;
	     case 0x36:
	     com.pref_seg=_ss_;
	     goto notdone;
	     case 0x3e:
	     com.pref_seg=_ds_;
	     goto notdone;
	     case 0x64:
	     com.pref_seg=_fs_;
	     goto notdone;
	     case 0x65:
	     com.pref_seg=_gs_;
	     goto notdone;
	     case 0xf0:
	     com.pref_lock=1;
	     goto notdone;
	     case 0xf2:
	     com.pref_rep=REPNE_NUM;
	     goto notdone;
	     case 0xf3:
	     com.pref_rep=REP_NUM;
	     goto notdone;
    	     }
com.op=&dis386[s];
#if 0
    if(com.op->gateway){
             (this->*com.op->decode)();
    }else if(com.op->haveMod)
		loadmodrm();
#else
//looks like this way works faster
    if(s==0x0f){
		com.op=&dis386_twobyte[nextbyte()];
		    if((com.op->op[0]&_gmask)==_gregGate){
    		     loadmodrm();
		     com.op=&(com.op->gateway[(com.modRM>>3)&0x7]);
		     }else
			if(com.op->haveMod)
		    	    loadmodrm();
	     }else if((com.op->op[0]&_gmask)==_gregGate){
	     loadmodrm();
	     com.op=&(com.op->gateway[(com.modRM>>3)&0x7]);
	     }else if(com.op->haveMod)	// one byte
		loadmodrm();

#endif		
    for(int i=0;i<3 /*&& com.op->opSize[i][0]*/;i++)
      com.opSize[i]=com.op->opSize[i][com.osize];
}


void disasm::print_str(int i){
      if(com.op->op[i]&0xf)
      	disout+=sprintf(disout,"es:%sdi",com.osize?"e":"");
      else
	disout+=sprintf(disout,"ds:%ssi",com.osize?"e":"");
}

void disasm::print_ptr(int i){
       disout+=sprintf(disout,"0x%x:0x%lx",com.segment,com.operand[i]);
}

void disasm::print_mx(int i){
      disout+=sprintf(disout,"[0x%lx]",com.operand[i]);
}

void disasm::print_rel(int i){
disout+=sprintf(disout,"%s",doSign(com.operand[i],com.opSize[i],S_MINUS));
}

void disasm::print_imm(int i){
      switch(com.opSize[i]){
        case 1:
	disout+=sprintf(disout,"byte 0x%02lx",com.operand[i]);
	return;
	case 2:
	disout+=sprintf(disout,"word 0x%04lx",com.operand[i]);
	return;
	case 4:
	disout+=sprintf(disout,"dword 0x%08lx",com.operand[i]);
	return;	
	}
}

void disasm::print_hsreg(int i){
disout+=sprintf(disout,"%s",_segreg_names[com.operand[i]]);
}

void disasm::print_hreg(int i){
   switch(com.opSize[i]){
   case 1:
     disout+=sprintf(disout,"%s",_reg_names[com.operand[i]].reg[0]);
   return;
   case 2:
     disout+=sprintf(disout,"%s",_reg_names[com.operand[i]].reg[1]);
   return;
   case 4:
     disout+=sprintf(disout,"%s",_reg_names[com.operand[i]].reg[2]);
   return;
   }
}
void disasm::print_dreg(int i){
        disout+=sprintf(disout,"dr%ld",com.operand[i]);
}
void disasm::print_creg(int i){
        disout+=sprintf(disout,"cr%ld",com.operand[i]);
}

void disasm::print_rm(int i){
char off[20]="";
    if(modrm16[com.modRM].rm_isreg)
	{
	   switch(com.opSize[i]){
	   case 1:
    	     disout+=sprintf(disout,"%s",_reg_names[com.base].reg[0]);
  	   return;
	   case 2:
	     disout+=sprintf(disout,"%s",_reg_names[com.base].reg[1]);
	   return;
           case 4:
       	     disout+=sprintf(disout,"%s",_reg_names[com.base].reg[2]);
	   return;
	   }
	}
      else{
           if(com.disp)
            sprintf(off,"%s",
	    doSign(com.operand[i],com.disp,
	    (com.base==_zero_ && com.index==_zero_ )?S_MINUS:S_ALL));
  	 disout+=sprintf(disout,"[");
	 
	   if(com.base!=_zero_)
	     disout+=sprintf(disout,"%s+",
	    	    _reg_names[com.base].reg[com.amode+1]);
		    
	   if(com.index!=_zero_){
		disout+=sprintf(disout,"%s",
		_reg_names[com.index].reg[com.amode+1]);
		
		    if(com.scale>1)
		    	disout+=sprintf(disout,"*%d",com.scale);
			
		disout+=sprintf(disout,"+");
		}
	   if(!(com.base==_zero_ && com.index==_zero_ ))
	     disout--;
	     
	   if(com.disp)
	     disout+=sprintf(disout,"%s",off);
	     
	 disout+=sprintf(disout,"]");
	 }
}

