/*
*  This file is part of Emma Virtual Machine
*  Copyright (C) 2003 Vitja Makarov ?<vitja@mail.ru>?
*  
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef _OPERATIONS_H
#define _OPERATIONS_H

#include "config.h"
// register type
const Bit32u _rcommon= 0x00100000;
const Bit32u _rseg   = 0x00200000;
const Bit32u _rctrl  = 0x00300000;
const Bit32u _rdbg   = 0x00400000;
const Bit32u _rFPU   = 0x00500000;
const Bit32u _rMMX   = 0x00600000;
const Bit32u _rSIMD  = 0x00700000;
const Bit32u _regMask= 0x00700000;

// operand size, these flags determinate
// real operand size
const Bit32u _oxMask=0xff000000;
const Bit32u _o8=    0x01000000;
const Bit32u _o16=   0x02000000;
const Bit32u _o32=   0x04000000;
const Bit32u _o48=   0x08000000;
const Bit32u _o64=   0x10000000;
const Bit32u _o128=  0x20000000;
// Pentium VI registers
const Bit32u _o256=  0x40000000;
// imm should be sign extended
const Bit32u _gsign= 0x80000000;

const Bit32u _haveMRM=0x8000;//
const Bit32u _haveNb=0x10000;// n bytes defined by _gXX flags
			     // follows after something like SIB
			     // MODR/M and it's offset
const Bit32u _gmask=0x7f00; 
const Bit32u _gno=0;
const Bit32u _gfirst=0x100;
//rel group
const Bit32u _grel=_gfirst*1;

const Bit32u _rel8=           _o8|_grel|_haveNb;
const Bit32u _rel16=         _o16|_grel|_haveNb;
const Bit32u _rel32=         _o32|_grel|_haveNb;
//const Bit32u _grelMask=_rel8|_rel16|_rel32;

//second group reg
const Bit32u _greg=_gfirst*2;//

// registers
const Bit32u _r8 =  _o8|_greg|_rcommon|_haveMRM;
const Bit32u _r16= _o16|_greg|_rcommon|_haveMRM;
const Bit32u _r32= _o32|_greg|_rcommon|_haveMRM;
const Bit32u _regctrl= _o32|_greg|_rctrl|_haveMRM;
const Bit32u _regdbg = _o32|_greg|_rdbg |_haveMRM;
const Bit32u _regseg = _o16|_greg|_rseg |_haveMRM;

// group rm
const Bit32u _grm=_gfirst*3;

const Bit32u _rm8 = _o8|_grm|_haveMRM;//
const Bit32u _rm16=_o16|_grm|_haveMRM;//
const Bit32u _rm32=_o32|_grm|_haveMRM;//
const Bit32u _indir=0x1;
// group imm
const Bit32u _gimm=_gfirst*4;
//const Bit32u _immMask=0x807;
const Bit32u _imm8 =  _o8|_gimm|_haveNb;//
const Bit32u _imm16= _o16|_gimm|_haveNb;//
const Bit32u _imm32= _o32|_gimm|_haveNb;//
const Bit32u _imm48= _o48|_gimm|_haveNb;//
//const Bit32u _gimmPre
// group mx
const Bit32u _gmx=_gfirst*5; 
//const Bit32u _gmxMask=0x100f;
const Bit32u _m=         _gmx;
const Bit32u _m8=    _o8|_gmx;//
const Bit32u _m16=  _o16|_gmx;//
const Bit32u _m32=  _o32|_gmx;//

//argument is hard-opcode-defined register
const Bit32u _ghreg=  _gfirst*6;
const Bit32u _reg8=  _o8|_ghreg|_rcommon;
const Bit32u _reg16=_o16|_ghreg|_rcommon;
const Bit32u _reg32=_o32|_ghreg|_rcommon;
const Bit32u _sreg= _o16|_ghreg|_rseg;
const Bit32u _creg= _o32|_ghreg|_rctrl;
const Bit32u _dreg= _o32|_ghreg|_rdbg;

const Bit32u _ral=0|_reg8;
const Bit32u _rcl=1|_reg8;
const Bit32u _rdl=2|_reg8;
const Bit32u _rbl=3|_reg8;
const Bit32u _rah=4|_reg8;
const Bit32u _rch=5|_reg8;
const Bit32u _rdh=6|_reg8;
const Bit32u _rbh=7|_reg8;

const Bit32u _rax=0|_reg16;
const Bit32u _rcx=1|_reg16;
const Bit32u _rdx=2|_reg16;
const Bit32u _rbx=3|_reg16;
const Bit32u _rsp=4|_reg16;
const Bit32u _rbp=5|_reg16;
const Bit32u _rsi=6|_reg16;
const Bit32u _rdi=7|_reg16;

const Bit32u _reax=0|_reg32;
const Bit32u _recx=1|_reg32;
const Bit32u _redx=2|_reg32;
const Bit32u _rebx=3|_reg32;
const Bit32u _resp=4|_reg32;
const Bit32u _rebp=5|_reg32;
const Bit32u _resi=6|_reg32;
const Bit32u _redi=7|_reg32;

const Bit32u _res=0x0|_sreg;
const Bit32u _rcs=0x1|_sreg;
const Bit32u _rss=0x2|_sreg;
const Bit32u _rds=0x3|_sreg;
const Bit32u _rfs=0x4|_sreg;
const Bit32u _rgs=0x5|_sreg;

const Bit32u _gptr=_gfirst*7;
const Bit32u _ptr=_o48|_o32|_grel|_haveNb;

const Bit32u _gstr=_gfirst*8;

const Bit32u _rdssi=_gstr|0;
const Bit32u _rdsdi=_gstr|1;

const Bit32u _gregGate=_gfirst*10; //mean's that command determinate's
const Bit32u _gtwobyte=_gfirst*11; //mean's that command determinate's

const Bit32u _regGate=_gregGate|_haveMRM;
#define IS_OP_REL(d) (((d)&_gmask)==_grel)
#define IS_OP_IMM(d) (((d)&_gmask)==_gimm)
#define IS_OP_RM(d)  (((d)&_gmask)==_grm)
#define IS_OP_indirRM(d)  ((((d)&_gmask)==_grm)&&((d)&_indir))
#define IS_OP_PTR(d)  (((d)&_gmask)==_gptr)

#endif
