#include <sys/vm86.h>
#include <signal.h>
#include "cpu.h"
#include "../debug/debug.h"

#define REGS  vm86s.regs

extern "C" int run_emma_loop();
extern "C" int emma_pseudo_vm86(void);

extern struct vm86plus_struct vm86s;
typedef Bit32u ioport_t;

extern volatile sig_atomic_t signal_pending;
volatile int in_debug=1;
volatile int int_pending=0;

extern "C" {
    extern Bit8u   port_inb(ioport_t port);
    extern Bit16u  port_inw(ioport_t port);
    extern Bit32u  port_ind(ioport_t port);
    extern void    port_outb(ioport_t port, Bit8u byte);
    extern void    port_outw(ioport_t port, Bit16u word);
    extern void    port_outd(ioport_t port, Bit32u word);
};

// syncronize registers up to EmmaCPU
void CPU::sync_regs_up(){
    _EAX=REGS.eax;
    _EBX=REGS.ebx;
    _ECX=REGS.ecx;
    _EDX=REGS.edx;
    _ESI=REGS.esi;
    _EBP=REGS.ebp;
    _ESP=REGS.esp;
#undef eip
    STH_THIS EIP=REGS.eip;

    _DS=REGS.ds;
    _CS=REGS.cs;
    _SS=REGS.ss;
    _ES=REGS.es;
    _FS=REGS.fs;
    _GS=REGS.gs;
    __eflags=REGS.eflags;
    
    sc_bit(__eflags, IF ,REGS.eflags & VIF);
    
}
// down to our father dosemu
void CPU::sync_regs_down(){
    REGS.eax=_EAX;
    REGS.ebx=_EBX;
    REGS.ecx=_ECX;
    REGS.edx=_EDX;
    REGS.esi=_ESI;
    REGS.ebp=_EBP;
    REGS.esp=_ESP;
    REGS.eip=STH_THIS EIP;

    REGS.ds=_DS;
    REGS.cs=_CS;
    REGS.ss=_SS;
    REGS.es=_ES;
    REGS.fs=_FS;
    REGS.gs=_GS;
    REGS.eflags=__eflags ; // | VIP;

    sc_bit(REGS.eflags, VIF ,__eflags & IF);

}

extern "C" void    run_caller_func ( int, unsigned char );

void CPU::emmaCtrl(){
char i;
/* we are in bios ... */
    if( _CS  < 0xf000 )
	{
	    i_int_cd_16();
	    return;
	}

i = * (char *) loadImm8 ();

STH_THIS sync_regs_down();

run_caller_func ( i, 1);

STH_THIS sync_regs_up();

}

static int in_hlt = 0;
static void hlt(void)
{
    in_hlt = 1;
}

extern "C" Bit8u   port_inb(ioport_t port);
extern "C" Bit16u  port_inw(ioport_t port);
extern "C" Bit32u  port_ind(ioport_t port);
extern "C" void    port_outb(ioport_t port, Bit8u byte);
extern "C" void    port_outw(ioport_t port, Bit16u word);
extern "C" void    port_outd(ioport_t port, Bit32u word);



CPU *cpu=NULL;

#if defined (_STH_)
CPU __cpu__;
#endif

extern "C" void init_emma_cpu();
void init_emma_cpu(){
#if defined (DEBUGGER)
    in_debug=1;
#else
    in_debug=0;
#endif

#if defined (_STH_)
    cpu=&__cpu__;
#else
    cpu=new CPU;
#endif


    cpu->set_memory((void *)0x0);
    cpu->sync_regs_up();

    /* int */
    cpu->set_handler ( &CPU::emmaCtrl, &CPU::emmaCtrl, OPCODE_ONEBYTE
							| 0xcd );
    /* hlt */
    cpu->set_handler ( &hlt, &hlt, OPCODE_ONEBYTE
							| 0xf4 );

}

void CPU::e_out8(Bit32u port, Bit8u src)
{
	port_outb(port, src);
}

void CPU::e_out16(Bit32u port, Bit16u src)
{
	port_outw(port, src);
}

void CPU::e_out32(Bit32u port, Bit32u src)
{
	port_outd(port, src);
}

Bit8u  CPU::e_in8 (Bit32u port)
{
	    return port_inb(port);
}

Bit16u CPU::e_in16(Bit32u port)
{
	    return port_inw(port);
}

Bit32u CPU::e_in32(Bit32u port)
{
	    return port_ind(port);
}


int CPU::external_check()
{
    if(in_hlt)
    {
	in_hlt = 0;
        cpu->sync_regs_down();
	REGS.eip--;
	return VM86_UNKNOWN;
    }


    if(signal_pending)
    {
        cpu->sync_regs_down();
	return VM86_SIGNAL;
    }
    
    if(vm86s.vm86plus.force_return_for_pic && get_bit(_EFLAGS, IF))
    {
        cpu->sync_regs_down();
	return VM86_PICRETURN;
    }
    
    return -1;    
}

int emma_pseudo_vm86(void)
{
    cpu->sync_regs_up();
    return cpu->run();
}
