/*
*  This file is part of Emma Virtual Machine
*  Copyright (C) 2003 Vitja Makarov ?<vitja@mail.ru>?
*  
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _HARDWARE_H_
#define _HARDWARE_H_
#include <sys/types.h>

#define SECTOR_SIZE 512

//bios disk error codes
#define BDE_NO      0x0 //00H  no error on last operation  
#define BDE_BADCMD  0x1 //01H  bad command: invalid request to controller                          
#define BDE_BADADR  0x2 //02H  bad address mark
#define BDE_WRPROT  0x3 //03H  write protect: attempted to write on write-protected diskette       
#define BDE_BADSEC  0x4 //04H  sector ID bad or not found
//06H  diskette change line is active
#define BDE_DMAFAIL 0x8  //08H  DMA failure                                                         
//09H  DMA overrun: attempted to write across a 64K-byte boundary.
#define BDE_NOMEDIA 0xc //0cH  Media type not available
//10H  bad CRC: Cyclical Redundancy Code does not agree with data          
//20H  diskette controller failure
//40H  bad seek; requested track not found                                 
//80H  time-out

/* int13 error returns */
#define DERR_NOERR	0
#define DERR_BADCMD 	1
#define DERR_BADSEC 	2
#define DERR_WP 	3
#define DERR_NOTFOUND 	4
#define DERR_CHANGE 	6
#define DERR_ECCERR 	0x10
#define DERR_CONTROLLER	0x20
#define DERR_SEEK	0x40
#define DERR_NOTREADY	0x80
#define DERR_WRITEFLT	0xcc


#ifdef __linux__
#define DISK_OFFSET(dp,h,s,t) \
  (((long)(t * dp->heads + h) * dp->sectors + s) * SECTOR_SIZE)
#else
#define DISK_OFFSET(dp,h,s,t) \
  (((t * dp->heads + h) * dp->sectors + s) * SECTOR_SIZE)
#endif

#ifndef __linux__
#define loff_t  off_t
#endif

/* disk file types */
typedef enum {
  NODISK = -1,
  IMAGE = 0, HDISK, FLOPPY, PARTITION, MAXIDX_DTYPES,
  NUM_DTYPES
} disk_t;

#define DISK_RDWR	0
#define DISK_RDONLY	1

/* definitions for 'dexeflags' in 'struct disk' and 'struct image_header' */
#define  DISK_IS_DEXE		1
#define  DISK_DEXE_RDWR		2

struct partition {
  int number;
  int beg_head, beg_sec, beg_cyl;
  int end_head, end_sec, end_cyl;
  long pre_secs;		/* sectors preceding partition */
  long num_secs;		/* sectors in partition */
  char *mbr;			/* fake Master Boot Record */
  int mbr_size;			/* usu. 1 sector */
};

struct disk {
  char *dev_name;		/* disk file */
  char *boot_name;              /* boot image file */
  int diskcyl4096;		/* INT13 support for 4096 cylinders */
  int wantrdonly;		/* user wants the disk to be read only */
  int rdonly;			/* The way we opened the disk (only filled in if the disk is open) */
  int dexeflags;		/* special flags for DEXE support */
  int sectors, heads, tracks;	/* geometry */
  long start;			/* geometry */
  int default_cmos;		/* default CMOS floppy type */
  disk_t type;			/* type of file: image, partition,
				   disk */
  loff_t header;		/* compensation for opt. pre-disk header */
  int fdesc;			/* below are not user settable */
  int removeable;		/* not user settable */
  int timeout;			/* seconds between floppy timeouts */
  struct partition part_info;	/* neato partition info */
};

class Hardware{
public:
Hardware();
int addFloppy(char *);
int readSector(int diskN,char *buffer,
	    long head,long sector,long track,long count);
struct disk* get_disk(int diskN);	    
struct disk fds[4];
int fdN;
int last_err;
};

#endif
