/*
 *  This file is part of Emma Virtual Machine
 *  Copyright (C) 2003 Vitja Makarov ?<vitja@mail.ru>?
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>

#include "disasm.h"
#include "debug/debug.h"
#include "cpu.h"
#include "hardware.h"

#if defined (_STH_)
CPU __cpu__;
#endif

int main(int argc,char **argv)
{
   FILE *fp;
   CPU *cpu;
   char floppy[100];
   int b=0;
   char *mem;
   extern Hardware *hw;
   Hardware hh;
   hw=&hh;
   
#if !defined (_STH_)
    cpu=new CPU();
#else
    cpu= &__cpu__;
#endif
   
    cpu->set_handler ( &CPU::emmaCtrl, &CPU::emmaCtrl, OPCODE_ONEBYTE
							| 0x82 );

   mem = (char *) malloc(1024 * 1024);
    if(!mem)
	{
	    fprintf(stderr,"Couldn't allocate memory!\n");
	    exit(-1);
	}
   
   cpu->set_memory(mem);
   
   signal(SIGINT, SIG_IGN);
   if(argc>1)
     strcpy(floppy,argv[1]);
   else
     strcpy(floppy,"bootss");
#if 0
   fp=fopen(floppy,"r");
   if(!fp)
     {
	perror("emma");
	exit(-1);
     }
#endif

#define GO_BIOS 1

#if GO_BIOS
   fp=fopen("bios/bios","r");
   if(!fp)
     perror("no bios found!!");
   else
     {
	b=fread(cpu->mem+cpu->lineaddr(0xf000,0),1,1024*64,fp);
	fclose(fp);
     }
   cpu->EIP=0xfff0;
   cpu->SRegs[_cs_]=0xf000;
//   cpu->stack_push(0,4);
#else
   fp=fopen(floppy,"r");
   if(!fp)
     {
	perror("sorry");
	exit(-1);
     }
   else
     {
	b=fread(cpu->mem+cpu->lineaddr(0,0x7c00),1,1024*64,fp);
	fclose(fp);
     }

   cpu->EIP=0x7c00;
   cpu->_CS=0x0;

#endif
   fprintf(stderr,"%d bytes written in 0xf000:0x0000\n",b);
   hw->addFloppy(floppy);
   cpu->run();
   return 0;
}

