/****************************************************************************
*                                                                           *
*                            Third Year Project                             *
*                                                                           *
*                            An IBM PC Emulator                             *
*                          For Unix and X Windows                           *
*                                                                           *
*                             By David Hedley                               *
*                                                                           *
*                                                                           *
* This program is Copyrighted.  Consult the file COPYRIGHT for more details *
*                                                                           *
****************************************************************************/

/* This is MAIN.C  This is where everything begins... */


#include "global.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>

#include "bios.h"
#include "vga.h"
#include "xstuff.h"
#include "hardware.h"
#include "vgahard.h"

#include "cpu.h"

BYTE *memory;
char *progname;

extern void run_emma_loop();
#ifndef DEBUGGER 
volatile int in_debug;
#endif

WORD wregs[8];	     /* Always little-endian */
BYTE *bregs[16];     /* Points to bytes within wregs[] */
unsigned sregs[6];   /* Always native machine word order */

unsigned ip;	     /* Always native machine word order */


    /* All the byte flags will either be 1 or 0 */

BYTE CF, PF, ZF, TF, IF, DF;

    /* All the word flags may be either none-zero (true) or zero (false) */

unsigned AF, OF, SF;


static UINT8 parity_table[256];
static BYTE *ModRMRegB[256];
static WORD *ModRMRegW[256];

    /* 'Shadows' of the segment registers. Point directly into the start of the
       segment within memory[] */
BYTE *c_cs,*c_ds,*c_es,*c_ss,*c_stack;

volatile int int_pending;   /* Interrupt pending */
volatile int int_blocked;   /* Blocked pending */

static struct
{
    BYTE **segment;
    WORD *reg1;
    WORD *reg2;
    int offset;
    int offset16;
    int pad[3];	/* Make structure power of 2 bytes wide for speed */
} ModRMRM[256];

static WORD zero_word = 0;

static WORD *ModRMRMWRegs[256];
static BYTE *ModRMRMBRegs[256];

void trap(void);


int emmavm;

void exit_emu(void)
{
    disable();
    bios_off();
    vga_off();
    exit(0);
}


void check_error(char *msg, int line)
{
    if (msg)
        fprintf(stderr,"%s, line %d\n", msg, line);
}


void read_pcemurc(void)
{         /* This procedure is a bit of a hack :) - will be redone later */
    FILE *f1;
    char buffer[1024];  /* Maximum path length. Should really be a #define */
    char keyword[1024];
    char value[1024];
    int line = 0;
    int tmp;

    /* Try current directory first... */

    if ((f1 = fopen(".pcemurc","r")) == NULL)
    {           /* Try home directory */
        sprintf(buffer,"%s/.pcemurc", getenv("HOME"));
        if ((f1 = fopen(buffer,"r")) == NULL)
        {
            printf("Warning: .pcemurc not found - using compile time defaults\n");
            return;
        }
    }

    while (fgets(buffer,sizeof buffer,f1) != NULL)
    {
        line++;
        tmp = sscanf(buffer," %s %s", keyword, value);

        if (tmp == 0 || keyword[0] == '#' || !buffer[0])
            continue;
        if (tmp != 2)
        {
            check_error("Syntax error in .pcemu file", line);
            continue;
        }

        if (strcasecmp(keyword,"bootfile") == 0)
            check_error(set_boot_file(value), line);
        else if (strcasecmp(keyword,"boottype") == 0)
            check_error(set_boot_type(strtol(value, NULL,10)), line);
        else if (strcasecmp(keyword,"updatespeed") == 0)
            check_error(set_update_rate(strtol(value, NULL,10)), line);
        else if (strcasecmp(keyword,"cursorspeed") == 0)
            check_error(set_cursor_rate(strtol(value, NULL,10)), line);
#ifdef EMMAVM	    
        else if (strcasecmp(keyword,"emmavm") == 0)
            emmavm=strtol(value, NULL,10)?1:0;
#endif	    
        else
            check_error("Syntax error in .pcemu file", line);
    }
    fclose(f1);
}           


void main(int argc, char **argv)
{
    progname = (progname = strrchr(argv[0],'/')) ? progname : argv[0];

#ifndef BOOT    
    FILE *f1;
    if (argc != 2)
    {
        fprintf(stderr,"Format: %s testfile\n",progname);
        exit(1);
    }
#endif
    if (!(memory = (BYTE *)malloc(MEMORY_SIZE)))
    {
        fprintf(stderr,"Insufficient available memory\n");
        exit(1);
    }
#ifdef BOOT
    memset(memory,0, MEMORY_SIZE);
#else
    if (!(f1 = fopen(argv[1],"rb")))
    {
        fprintf(stderr,"Cannot open test file\n");
        exit(1);
    }


    fread(memory+0x800,1,MEMORY_SIZE-0x800,f1);
    fclose(f1);
#endif

    read_pcemurc();
    disable();

    start_X();
    init_cpu();
    init_bios();
    init_vga();
//	if(emmavm)
//	    in_debug=1;
//    if(!emmavm)
//	init_timer();
    
    enable();
// some os'es need this, and all bios'es know it, except our
    wregs[SP]=0x100;
    sregs[SS]=0x40;
    sregs[DS]=0x40;

    c_cs = SegToMemPtr(CS);
    c_ds = SegToMemPtr(DS);
    c_es = SegToMemPtr(ES);
    c_stack = c_ss = SegToMemPtr(SS);
    
//    
    
//      if(!emmavm)
//        execute();
//      else
#if 1
    	init_timer();
        run_emma_loop();
#else	
    void diff_cpu();
    extern void diff_cpu();    
    diff_cpu();
#endif    

    /* NOT REACHED */
}


#if 0
void init_cpu(void)
{
    unsigned int i,j,c;
    
    for (i = 0; i < 4; i++)
    {
        wregs[i] = 0;
        sregs[i] = 0x70;
    }
    for (; i < 8; i++)
        wregs[i] = 0;
    
    wregs[SP] = 0;
    ip = 0x100;
    
    for (i = 0;i < 256; i++)
    {
        for (j = i, c = 0; j > 0; j >>= 1)
            if (j & 1) c++;
        
        parity_table[i] = !(c & 1);
    }
    
    CF = PF = AF = ZF = SF = TF = IF = DF = OF = 0;
    
    bregs[AL] = (BYTE *)&wregs[AX];
    bregs[AH] = (BYTE *)&wregs[AX]+1;
    bregs[CL] = (BYTE *)&wregs[CX];
    bregs[CH] = (BYTE *)&wregs[CX]+1;
    bregs[DL] = (BYTE *)&wregs[DX];
    bregs[DH] = (BYTE *)&wregs[DX]+1;
    bregs[BL] = (BYTE *)&wregs[BX];
    bregs[BH] = (BYTE *)&wregs[BX]+1;
    
    bregs[SPL] = (BYTE *)&wregs[SP];
    bregs[SPH] = (BYTE *)&wregs[SP]+1;
    bregs[BPL] = (BYTE *)&wregs[BP];
    bregs[BPH] = (BYTE *)&wregs[BP]+1;
    bregs[SIL] = (BYTE *)&wregs[SI];
    bregs[SIH] = (BYTE *)&wregs[SI]+1;
    bregs[DIL] = (BYTE *)&wregs[DI];
    bregs[DIH] = (BYTE *)&wregs[DI]+1;
    
    for (i = 0; i < 256; i++)
    {
        ModRMRegB[i] = bregs[(i & 0x38) >> 3];
        ModRMRegW[i] = &wregs[(i & 0x38) >> 3];
    }
    
    for (i = 0; i < 0x40; i++)
    {
        switch (i & 7)
        {
        case 0:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &wregs[BX];
            ModRMRM[i].reg2 = &wregs[SI];
            ModRMRM[i].offset = ModRMRM[i].offset16 = 0;
            break;
        case 1:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &wregs[BX];
            ModRMRM[i].reg2 = &wregs[DI];
            ModRMRM[i].offset = ModRMRM[i].offset16 = 0;
            break;
        case 2:
            ModRMRM[i].segment = &c_ss;
            ModRMRM[i].reg1 = &wregs[BP];
            ModRMRM[i].reg2 = &wregs[SI];
            ModRMRM[i].offset = ModRMRM[i].offset16 = 0;
            break;
        case 3:
            ModRMRM[i].segment = &c_ss;
            ModRMRM[i].reg1 = &wregs[BP];
            ModRMRM[i].reg2 = &wregs[DI];
            ModRMRM[i].offset = ModRMRM[i].offset16 = 0;
            break;
        case 4:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &zero_word;
            ModRMRM[i].reg2 = &wregs[SI];
            ModRMRM[i].offset = ModRMRM[i].offset16 = 0;
            break;
        case 5:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &zero_word;
            ModRMRM[i].reg2 = &wregs[DI];
            ModRMRM[i].offset = ModRMRM[i].offset16 = 0;
            break;
        case 6:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &zero_word;
            ModRMRM[i].reg2 = &zero_word;
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 1;
            break;
        default:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &wregs[BX];
            ModRMRM[i].reg2 = &zero_word;
            ModRMRM[i].offset = ModRMRM[i].offset16 = 0;
            break;
        }
    }
    
    for (i = 0x40; i < 0x80; i++)
    {
        switch (i & 7)
        {
        case 0:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &wregs[BX];
            ModRMRM[i].reg2 = &wregs[SI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 0;
            break;
        case 1:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &wregs[BX];
            ModRMRM[i].reg2 = &wregs[DI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 0;
            break;
        case 2:
            ModRMRM[i].segment = &c_ss;
            ModRMRM[i].reg1 = &wregs[BP];
            ModRMRM[i].reg2 = &wregs[SI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 0;
            break;
        case 3:
            ModRMRM[i].segment = &c_ss;
            ModRMRM[i].reg1 = &wregs[BP];
            ModRMRM[i].reg2 = &wregs[DI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 0;
            break;
        case 4:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &zero_word;
            ModRMRM[i].reg2 = &wregs[SI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 0;
            break;
        case 5:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &zero_word;
            ModRMRM[i].reg2 = &wregs[DI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 0;
            break;
        case 6:
            ModRMRM[i].segment = &c_ss;
            ModRMRM[i].reg1 = &wregs[BP];
            ModRMRM[i].reg2 = &zero_word;
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 0;
            break;
        default:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &wregs[BX];
            ModRMRM[i].reg2 = &zero_word;
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 0;
            break;
        }
    }
    
    for (i = 0x80; i < 0xc0; i++)
    {
        switch (i & 7)
        {
        case 0:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &wregs[BX];
            ModRMRM[i].reg2 = &wregs[SI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 1;
            break;
        case 1:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &wregs[BX];
            ModRMRM[i].reg2 = &wregs[DI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 1;
            break;
        case 2:
            ModRMRM[i].segment = &c_ss;
            ModRMRM[i].reg1 = &wregs[BP];
            ModRMRM[i].reg2 = &wregs[SI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 1;
            break;
        case 3:
            ModRMRM[i].segment = &c_ss;
            ModRMRM[i].reg1 = &wregs[BP];
            ModRMRM[i].reg2 = &wregs[DI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 1;
            break;
        case 4:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &zero_word;
            ModRMRM[i].reg2 = &wregs[SI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 1;
            break;
        case 5:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &zero_word;
            ModRMRM[i].reg2 = &wregs[DI];
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 1;
            break;
        case 6:
            ModRMRM[i].segment = &c_ss;
            ModRMRM[i].reg1 = &wregs[BP];
            ModRMRM[i].reg2 = &zero_word;
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 1;
            break;
        default:
            ModRMRM[i].segment = &c_ds;
            ModRMRM[i].reg1 = &wregs[BX];
            ModRMRM[i].reg2 = &zero_word;
            ModRMRM[i].offset = 1;
            ModRMRM[i].offset16 = 1;
            break;
        }
    }
    
    for (i = 0xc0; i < 0x100; i++)
    {
        ModRMRMWRegs[i] = &wregs[i & 7];
        ModRMRMBRegs[i] = bregs[i & 7];
    }
}
#endif