#include <sys/vm86.h>
#include "../cpu/cpu.h"
#include "../debug/debug.h"

#include "newinls.h"


extern volatile int int_pending;
extern "C" void external_int(void);
extern "C" void enable();
extern "C" void disable();


int CPU::external_check(void){

    if (int_pending && __eflags & IF )
    {
    disable();
    int OP_1=int_pending;
    STH_THIS amode=STH_THIS osize=0;
    STH_THIS segment=_zero_;

    void *int_ptr;
    stack_push16(__eflags);
    clear_bit(__eflags,IF|TF|AC);
    sync_eip();
    stack_push16(_CS);
    stack_push16(STH_THIS EIP);
    int_ptr=(void *)memptr(0,4*OP_1);
    STH_THIS EIP=*(Bit16u *)int_ptr;
    _CS=*(++(Bit16u *)int_ptr);
    sync_code();
    
    int_pending = 0;
    enable();
    }
    return -1;    
}


#undef CF
#undef PF
#undef ZF
#undef TF
#undef IF
#undef DF
#undef AF
#undef OF
#undef SF

extern "C" void run_emma_loop();

extern Bit16u wregs[8];
extern unsigned sregs[4];
extern Bit8u  *memory;
extern unsigned ip;
extern Bit8u *c_cs,*c_ds,*c_es,*c_ss,*c_stack;
extern volatile int in_debug;

extern Bit8u CF, PF, ZF, TF, IF, DF;
extern unsigned AF, OF, SF;


typedef void (*int_handler)(void);
int_handler int_handlers[0x100];
extern "C" void int9(void);
extern "C" void int_serial(void);
extern "C" void int_printer(void);
extern "C" void int_keyboard(void);
extern "C" void int_extended(void);
extern "C" void int_basic(void);
extern "C" void int_reboot(void);
extern "C" void int_equipment(void);
extern "C" void int_memory(void);
extern "C" void int_time(void);
extern "C" void int_disk(void);
extern "C" void int_e6(void);
extern "C" void int_2f(void);
extern "C" void int_video(void);
extern "C" void i_gobios(void);
extern "C" void sync_regs_up(void);
extern "C" void sync_regs_down(void);
extern "C" void write_port(unsigned,Bit8u);
extern "C" void  write_port(unsigned,Bit8u);
extern "C" Bit8u read_port(unsigned);

inline void init_handlers(){
    for(int i=0;i<0x100;i++)
	int_handlers[i]=NULL;
	
//int_handlers[0x8]=int_;
int_handlers[0x9]=int9;
int_handlers[0x10]=int_video;
int_handlers[0x11]=int_equipment;
int_handlers[0x12]=int_memory;
int_handlers[0x13]=int_disk;
int_handlers[0x14]=int_serial;
int_handlers[0x15]=int_extended;
int_handlers[0x16]=int_keyboard;
int_handlers[0x17]=int_printer;
//int_handlers[0x18]=int_basic;
int_handlers[0x19]=int_reboot;
int_handlers[0x1a]=int_time;
int_handlers[0xe6]=int_e6;
int_handlers[0xe8]=int_2f;
int_handlers[0x2f]=int_2f;

}


//extern  *c_cs,*c_ds,*c_es,*c_ss,*c_stack;
// syncronize registers up to EmmaCPU
void CPU::sync_regs_up(){
    for(int i=0;i<8;i++)
     LOWORD(STH_THIS Regs[i])=wregs[i];
    for(int i=0;i<4;i++)
     STH_THIS SRegs[i]=::sregs[i];
 sc_bit(__eflags,1<<0,CF);
 sc_bit(__eflags,1<<2,PF);
 sc_bit(__eflags,1<<6,ZF);
 sc_bit(__eflags,1<<8,TF);
 sc_bit(__eflags,1<<9,IF);
 sc_bit(__eflags,1<<10,DF);
 sc_bit(__eflags,1<<11,OF);
 sc_bit(__eflags,1<<4,AF);
 sc_bit(__eflags,1<<7,SF);
 STH_THIS EIP=ip;
}
// down to our father dosemu
void CPU::sync_regs_down(){
    for(int i=0;i<8;i++)
     wregs[i]=LOWORD(STH_THIS Regs[i]);
    for(int i=0;i<4;i++)
     ::sregs[i]=STH_THIS SRegs[i];
 CF=get_bit(STH_THIS eflags,1<<0);
 PF=get_bit(STH_THIS eflags,1<<2);
 ZF=get_bit(STH_THIS eflags,1<<6);
 TF=get_bit(STH_THIS eflags,1<<8);
 IF=get_bit(STH_THIS eflags,1<<9);
 DF=get_bit(STH_THIS eflags,1<<10);
 OF=get_bit(STH_THIS eflags,1<<11);
 AF=get_bit(STH_THIS eflags,1<<4);
 SF=get_bit(STH_THIS eflags,1<<7);
ip=STH_THIS EIP;
c_cs=(Bit8u*)memptr(_CS,0);
c_ds=(Bit8u*)memptr(_DS,0);
c_es=(Bit8u*)memptr(_ES,0);
c_ss=(Bit8u*)memptr(_SS,0);
c_stack=(Bit8u*)memptr(_SS,_SP);
}

void
 CPU::emmaCtrl()
 {
    char i;
    
    i = * (char *) loadImm8();
STH_THIS sync_regs_down();

    if(int_handlers[i])
	int_handlers[i]();
    else
        fprintf(stderr,"Unhandled bios interrupt: 0x%x\n",i);
	    
STH_THIS sync_regs_up();
}

#define DD_IPORT(d,s) Bit##s##u CPU::e_##d##s(Bit32u port){ \
    	return read_port(port);	\
    }

#define DD_OPORT(d,s) void CPU::e_##d##s(Bit32u port,Bit##s##u src){ \
	write_port(port,src);	\
    }

DD_IPORT(in,8)
DD_IPORT(in,16)
DD_IPORT(in,32)
DD_OPORT(out,8)
DD_OPORT(out,16)
DD_OPORT(out,32)

CPU *cpu=NULL;
CPU __cpu__;


void init_emma_cpu(){

#if defined(DEBUGGER)
    in_debug=1;
#else
    in_debug=0;
#endif
#if defined (_STH_)
    cpu=&__cpu__;
#else
    cpu=new CPU();
#endif
cpu->set_memory((void *)memory);
cpu->sync_regs_up();
cpu->eflags|=0xf000;

init_handlers();

cpu->set_handler ( &CPU::emmaCtrl, &CPU::emmaCtrl, OPCODE_ONEBYTE
							| 0x82 );

}



void run_emma_loop(){
init_emma_cpu();
cpu->run();
}

void sync_regs_up(){
    if(cpu)
	cpu->sync_regs_up();
}


void sync_regs_down(){
    if(cpu)
	cpu->sync_regs_down();
}
